/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996 - 1998, Emmanuel CURIS
   Cration de ce fichier : 22 septembre 1998

   Extraction du signal EXAFS proprement dit

   Normalisation du signal EXAFS
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "SYSTEME.H"

#include "ERREURS.H"
#include "EXAFS.H"
#include "GRAPHES.H"

#include "ATOMES.H"
#include "EXTRAIT.H"
#include "LISSAGE.H"
#include "SCIENCE.H"

/* ~~~~~~~~ Variables locales  ce fichier ~~~~~~~~ */

#define LG_MAX_NOM_GRAPHE 50 /* caractres au plus pour le nom du graphe mu_0 dans l'arbre */

static Graphe *mu0_normalise = NULL;

static int methode = MU0_LANSBERGER; /* Mthode thorique de normalisation */

static const double k_v_E = H_BARRE * H_BARRE * M_V_ANGSTROM * M_V_ANGSTROM / (2 * MASSE_ELECTRON * EV_V_JOULE);

/* ~~~~~~~~ Fonctions locales  ce fichier ~~~~~~~~ */

static boolean gere_normalise   (int objet, int x, int y);
static Graphe *prepare_normalise(Graphe *source, boolean nouveau);
static void    propose_valeurs  (void);
static double  mu0_mcmaster     (double e, double a0, double a1, double a2, double a3);

/* ~~~~~~~~ Ralisation pratique du module ~~~~~~~~ */

/* -------- Interface de la normalisation ----------\
 | Renvoi le graphe normalis (NULL si erreur)      |
 |                                                  |
 | source : le graphe  normaliser                  |
 \------------------------------------------------- */

Graphe *norm_exafs(Graphe *source)
{
  int retour;
  Graphe *graphe;
  boolean mu0_present = FAUX;
  double e_0; /* nergie de seuil, pour les modles thoriques */

  if (source == NULL) return NULL; /* Pas de graphe en entre */
  if ( source->drapeaux.complexe != GR_NORMAL )
  {
    EXF_ERREUR(GRAPHE_COMPLEXE);
    return NULL;
  }

  for (graphe = premier_graphe ; graphe ; graphe = graphe->suivant)
  {
    if ( (graphe->type == GR_SSTYPE) || (graphe->type == GR_MU0) )
    {
      mu0_present = VRAI;
      break;
    }
  }

  mu0_normalise = NULL;
  OBJ_active(NORM_EXAFS, MU0_CALCULE, mu0_present);
  OBJ_active(NORM_EXAFS, MU0_A_UTILISER, mu0_present);
  if (mu0_present == FAUX)
  {
    OBJ_choisit(NORM_EXAFS, MU0_THEORIQUE, FORCE_OUI);
#ifdef _FRANCAIS_
    met_chaine(NORM_EXAFS, MU0_A_UTILISER, LG_MAX_NOM_GRAPHE,
	       "Aucun graphe de ce type en m"e_AIGU"moire", TXT_COPIER);
#endif
#ifdef _ENGLISH_
    met_chaine(NORM_EXAFS, MU0_A_UTILISER, LG_MAX_NOM_GRAPHE,
	       "No suitable graph in memory", TXT_COPIER);
#endif
    OBJ_choisit(NORM_EXAFS, methode, FORCE_OUI);
  }
  else
  {
    OBJ_active(NORM_EXAFS, MU0_A_UTILISER, OBJ_choisit(NORM_EXAFS, MU0_CALCULE, QUESTION) );
#ifdef _FRANCAIS_
    met_chaine(NORM_EXAFS, MU0_A_UTILISER, LG_MAX_NOM_GRAPHE,
	       "Aucun graphe s"e_AIGU"lectionn"e_AIGU, TXT_COPIER);
#endif
#ifdef _ENGLISH_
    met_chaine(NORM_EXAFS, MU0_A_UTILISER, LG_MAX_NOM_GRAPHE,
	       "No selected graph", TXT_COPIER);
#endif
  }

  arbres[NORM_EXAFS].fonction.arbre = gere_normalise;

  do
  {
    retour = ext_formdo(NORM_EXAFS);
    if (retour == NORM_ANNULER)
      return NULL;
  } while (retour != NORMALISER);

  if ( OBJ_choisit(NORM_EXAFS, MU0_CALCULE, QUESTION) ) /* On utilise un mu_0 calcul par ailleurs */
  {
    double *b, *c, *d;
    double retour;
    unsigned long i;
    register double *x_lu, *y_lu, *y_norm;

    if (mu0_normalise == NULL)
    {
      printf("Graphe de normalisation non dfini\n");
      return NULL;
    }

    if ( ( mu0_normalise->x_max <= source->x_min )
      || ( mu0_normalise->x_min >= source->x_max ) )
    {
      EXF_ERREUR(ERR_DISJOINTS);
      return NULL;
    }

    retour = SPL_tt_prep(mu0_normalise, &b, &c, &d);
    if (retour < 0)
      return NULL;

    graphe = prepare_normalise(source, OBJ_choisit(NORM_EXAFS, NORM_EXF_NOUVEAU, QUESTION) );
    if (graphe == NULL) /* Cration a chou */
      return NULL;

    for (i = 0, x_lu = source->x, y_lu = source->y, y_norm = graphe->y ;
	 i < graphe->nombre_point ;
	 i++, x_lu++, y_lu++, y_norm++)
    {
      double mu0 = SPL_tt_calc(mu0_normalise, *x_lu, b, c, d);
      if (mu0)
	*y_norm = *y_lu / mu0;
      else
      {
	EXF_ERREUR(DIV_ZERO);
	*y_norm = *y_lu; /* Normalisation impossible sur ce point... */
      }
    }

    free(b); free(c); free(d); /* Plus besoin d'eux */
    GRF_bornes(graphe, NE_PAS_FAIRE, DESORDONNE);
    return graphe;
  }

  /* Si on est l, c'est que l'on veut utiliser une formule thorique de mu_0 */

  if ( OBJ_choisit(NORM_EXAFS, MU0_CONSTANT, QUESTION) ) /* mu_0 = constante */
  {
    double mu0 = prend_fvaleur(NORM_EXAFS, NORM_CONSTANTE_1);
    unsigned long i;
    register double *x_lu, *y_lu, *y_norm;
    unsigned int j, degre = prend_valeur(NORM_EXAFS, NORM_DEGRE);

    if (mu0 == 0)
    {
      EXF_ERREUR(DIV_ZERO);
      return NULL;
    }

    graphe = prepare_normalise(source, OBJ_choisit(NORM_EXAFS, NORM_EXF_NOUVEAU, QUESTION) );
    if (graphe == NULL) /* Cration a chou */
      return NULL;

    for (i = 0, x_lu = source->x, y_lu = source->y, y_norm = graphe->y ;
	 i < graphe->nombre_point ;
	 i++, x_lu++, y_lu++, y_norm++)
    {
      if (degre)
      {
	double k = *x_lu;
	for (j = 1 ; j < degre ; j++)
	  k *= *x_lu;

	mu0 *= k; /* On multiplie par k^n */
      }
      *y_norm = *y_lu / mu0;
    }

    GRF_bornes(graphe, NE_PAS_FAIRE, DESORDONNE);
    return graphe;
  }    

  /* Si on est l, on utilise une formule variant avec E0 -> il faut lire e0 */

  e_0 = prend_fvaleur(NORM_EXAFS, NORM_CONSTANTE_1);
  if (e_0 == 0)
  {
    EXAFS_ERREUR(E0_INCORRECT);
    return NULL;
  }

  if ( OBJ_choisit(NORM_EXAFS, MU0_LANSBERGER, QUESTION) ) /* mu_0 = donn par Lengeler & Eisenberger */
  {
    unsigned long i;
    register double *x_lu, *y_lu, *y_norm;
    unsigned int j, degre = prend_valeur(NORM_EXAFS, NORM_DEGRE);

    graphe = prepare_normalise(source, OBJ_choisit(NORM_EXAFS, NORM_EXF_NOUVEAU, QUESTION) );
    if (graphe == NULL) /* Cration a chou */
      return NULL;
 
    if ( OBJ_choisit(NORM_EXAFS, NORM_EV, QUESTION) ) /* On travaille en chelle d'nergie */
    {
      if ( graphe->x_min < e_0 )
      {
	/* Alerte : le graphe commence _avant_ e0 */
      }

      for (i = 0, x_lu = source->x, y_lu = source->y, y_norm = graphe->y ;
	   i < graphe->nombre_point ;
	   i++, x_lu++, y_lu++, y_norm++)
      {
	double mu0 = 11.0/3.0 - 8.0/3.0 * *x_lu / e_0;

	if (mu0)
	  *y_norm = *y_lu / mu0;
	else
	{
	  EXF_ERREUR(DIV_ZERO);
	  *y_norm = *y_lu; /* Normalisation impossible sur ce point... */
	}
      }
    }
    else                                              /* On travaille en chelle des k */
    {
      double coef = 8 * k_v_E / (3 * e_0);

      for (i = 0, x_lu = source->x, y_lu = source->y, y_norm = graphe->y ;
	   i < graphe->nombre_point ;
	   i++, x_lu++, y_lu++, y_norm++)
      {
	double mu0 = 1 - coef * *x_lu * *x_lu;

	if (degre)
	{
	  double k = *x_lu;
	  for (j = 1 ; j < degre ; j++)
	    k *= *x_lu;

	  mu0 *= k; /* On multiplie par k^n */
	}

	if (mu0)
	  *y_norm = *y_lu / mu0;
	else
	{
	  EXF_ERREUR(DIV_ZERO);
	  *y_norm = *y_lu; /* Normalisation impossible sur ce point... */
	}
      }
    }

    GRF_bornes(graphe, NE_PAS_FAIRE, DESORDONNE);
    return graphe;
  }

  if ( OBJ_choisit(NORM_EXAFS, MU0_VICTOREEN, QUESTION) ) /* mu_0 = donn par Victoreen */
  {
    unsigned long i;
    register double *x_lu, *y_lu, *y_norm;
    double delta_c, delta_d;
    double coef;
    double facteur;
    unsigned int j, degre = prend_valeur(NORM_EXAFS, NORM_DEGRE);

    /* Delta_x = x(aprs le seuil) - x(avant le seuil) */

    delta_c = prend_fvaleur(NORM_EXAFS, NORM_CTE_4) - prend_fvaleur(NORM_EXAFS, NORM_CTE_2);
    delta_d = prend_fvaleur(NORM_EXAFS, NORM_CTE_5) - prend_fvaleur(NORM_EXAFS, NORM_CTE_3);

    facteur = PLANCK * VITESSE_LUMIERE * M_V_ANGSTROM / (e_0 * EV_V_JOULE);
    facteur = facteur * facteur * facteur * ( delta_c - delta_d * facteur); /* C(hc/E0) - D(hc/E0)^4 */
    if (facteur == 0)
    {
      EXF_ERREUR(DIV_ZERO);
      return NULL;
    }

    graphe = prepare_normalise(source, OBJ_choisit(NORM_EXAFS, NORM_EXF_NOUVEAU, QUESTION) );
    if (graphe == NULL) /* Cration a chou */
      return NULL;

    if ( OBJ_choisit(NORM_EXAFS, NORM_EV, QUESTION) ) /* On travaille en chelle d'nergie */
    {
      for (i = 0, x_lu = source->x, y_lu = source->y, y_norm = graphe->y ;
	   i < source->nombre_point ;
	   i++, x_lu++, y_lu++, y_norm++)
      {
	if ( *x_lu )
	{
	  EXF_ERREUR(DIV_ZERO);
	  *y_norm = 0;
	}
	else
	{
	  double lambda = PLANCK * VITESSE_LUMIERE * M_V_ANGSTROM / (*x_lu * EV_V_JOULE) ; /* lambda en angstroms... */
	  double l_cube = lambda * lambda * lambda;                        /* Cube de la longueur d'onde */
	  double mu_0;

	  mu_0 = l_cube * (delta_c - delta_d * lambda);
	  if (mu_0)
	    *y_norm = *y_lu / mu_0;
	  else
	  {
	    EXF_ERREUR(DIV_ZERO);
	    *y_norm = 0;
	  }
	}
      }
    }
    else
    {
      for (i = 0, x_lu = source->x, y_lu = source->y, y_norm = graphe->y ;
	   i < source->nombre_point ;
	   i++, x_lu++, y_lu++, y_norm++)
      {
	double nrj = e_0 + *x_lu * *x_lu * k_v_E; /* Valeur de l'nergie  partir de k */

	if (nrj == 0)
	{
	  EXF_ERREUR(DIV_ZERO);
	  *y_norm = 0;
	}
	else
	{
	  double lambda = PLANCK * VITESSE_LUMIERE * M_V_ANGSTROM / (nrj * EV_V_JOULE);
	  double mu_0 = lambda * lambda * lambda * (delta_c - delta_d * lambda);

	  if (degre)
	  {
	    double k = *x_lu;
	    for (j = 1 ; j < degre ; j++)
	      k *= *x_lu;
	    
	    mu_0 *= k; /* On multiplie par k^n */
	  }

	  if (mu_0)
	    *y_norm = *y_lu  * facteur / mu_0;
	  else
	  {
	    EXF_ERREUR(DIV_ZERO);
	    *y_norm = 0;
	  }
	}
      }
    }

    GRF_bornes(graphe, NE_PAS_FAIRE, DESORDONNE);
    return graphe;
  }

  if ( OBJ_choisit(NORM_EXAFS, MU0_MAC_KALE, QUESTION) ) /* mu_0 = donn par McMaster */
  {
    unsigned long i;
    register double *x_lu, *y_lu, *y_norm;
    double a0_av, a1_av, a2_av, a3_av, /* Coefficients avant le seuil */
           a0_ap, a1_ap, a2_ap, a3_ap; /* Coefficients aprs le seuil */
    double coef;
    double facteur;
    unsigned int j, degre = prend_valeur(NORM_EXAFS, NORM_DEGRE);

    a0_av = prend_fvaleur(NORM_EXAFS, NORM_CTE_2); /* Coefficients avant le seuil */
    a1_av = prend_fvaleur(NORM_EXAFS, NORM_CTE_3);
    a2_av = prend_fvaleur(NORM_EXAFS, NORM_CTE_4);
    a3_av = prend_fvaleur(NORM_EXAFS, NORM_CTE_5);
    a0_ap = prend_fvaleur(NORM_EXAFS, NORM_CTE_6); /* Coefficients aprs le seuil */
    a1_ap = prend_fvaleur(NORM_EXAFS, NORM_CTE_7);
    a2_ap = prend_fvaleur(NORM_EXAFS, NORM_CTE_8);
    a3_ap = prend_fvaleur(NORM_EXAFS, NORM_CTE_9);

    facteur  = mu0_mcmaster(e_0, a0_ap, a1_ap, a2_ap, a3_ap);
    facteur -= mu0_mcmaster(e_0, a0_av, a1_av, a2_av, a3_av);

    if (facteur == 0)
    {
      EXF_ERREUR(DIV_ZERO);
      return NULL;
    }

    graphe = prepare_normalise(source, OBJ_choisit(NORM_EXAFS, NORM_EXF_NOUVEAU, QUESTION) );
    if (graphe == NULL) /* Cration a chou */
      return NULL;

    if ( OBJ_choisit(NORM_EXAFS, NORM_EV, QUESTION) ) /* On travaille en chelle d'nergie */
    {
      for (i = 0, x_lu = source->x, y_lu = source->y, y_norm = graphe->y ;
	   i < source->nombre_point ;
	   i++, x_lu++, y_lu++, y_norm++)
      {
	double mu_0;

	mu_0  = mu0_mcmaster(*x_lu, a0_ap, a1_ap, a2_ap, a3_ap);
	mu_0 -= mu0_mcmaster(*x_lu, a0_av, a1_av, a2_av, a3_av);
	mu_0 /= facteur;

	if (mu_0)
	  *y_norm = *y_lu / mu_0;
	else
	{
	  EXF_ERREUR(DIV_ZERO);
	  *y_norm = 0;
	}
      }
    }
    else /* Travail en chelle d'nergie */
    {
      for (i = 0, x_lu = source->x, y_lu = source->y, y_norm = graphe->y ;
	   i < source->nombre_point ;
	   i++, x_lu++, y_lu++, y_norm++)
      {
	double nrj = e_0 + *x_lu * *x_lu * k_v_E; /* Valeur de l'nergie  partir de k */
	double mu_0;

	mu_0  = mu0_mcmaster(nrj, a0_ap, a1_ap, a2_ap, a3_ap);
	mu_0 -= mu0_mcmaster(nrj, a0_av, a1_av, a2_av, a3_av);
	mu_0 /= facteur;

	if (degre)
	{
	  double k = *x_lu;
	  for (j = 1 ; j < degre ; j++)
	    k *= *x_lu;
	    
	  mu_0 *= k; /* On multiplie par k^n */
	}

	if (mu_0)
	  *y_norm = *y_lu  * facteur / mu_0;
	else
	{
	  EXF_ERREUR(DIV_ZERO);
	  *y_norm = 0;
	}
      }
    }

    GRF_bornes(graphe, NE_PAS_FAIRE, DESORDONNE);
    return graphe;
  }

  /* Reste  faire la mthode de Heitler */

  return NULL;
}

/* -------- Gestion de l'interface de normalisation (procdure habituelle) -------- */

static boolean gere_normalise(int objet, int x, int y)
{
  switch (objet)
  {
  case MU0_CALCULE:
    OBJ_choisit(NORM_EXAFS, MU0_CALCULE, FORCE_OUI);
    OBJ_active (NORM_EXAFS, MU0_A_UTILISER, FORCE_OUI);
    OBJ_active (NORM_EXAFS, NORM_PROPOSER , FORCE_NON);
    OBJ_active (NORM_EXAFS, MU0_CONSTANT  , FORCE_NON);
    OBJ_active (NORM_EXAFS, MU0_HEITLER   , FORCE_NON);
    OBJ_active (NORM_EXAFS, MU0_LANSBERGER, FORCE_NON);
    OBJ_active (NORM_EXAFS, MU0_MAC_KALE  , FORCE_NON);
    OBJ_active (NORM_EXAFS, MU0_VICTOREEN , FORCE_NON);
    OBJ_active (NORM_EXAFS, NORM_CONSTANTE_1, FORCE_NON);
    OBJ_active (NORM_EXAFS, NORM_CTE_2, FORCE_NON);
    OBJ_active (NORM_EXAFS, NORM_CTE_3, FORCE_NON);
    OBJ_active (NORM_EXAFS, NORM_CTE_4, FORCE_NON);
    OBJ_active (NORM_EXAFS, NORM_CTE_5, FORCE_NON);
    OBJ_active (NORM_EXAFS, NORM_CTE_6, FORCE_NON);
    OBJ_active (NORM_EXAFS, NORM_CTE_7, FORCE_NON);
    OBJ_active (NORM_EXAFS, NORM_CTE_8, FORCE_NON);
    break;

  case MU0_A_UTILISER:
    mu0_normalise = GRF_choisir(GR_MU0);
    if (mu0_normalise)
      met_chaine(NORM_EXAFS, MU0_A_UTILISER, LG_MAX_NOM_GRAPHE,
		 mu0_normalise->nom, TXT_COPIER);
    else
    {
#ifdef _FRANCAIS_
    met_chaine(NORM_EXAFS, MU0_A_UTILISER, LG_MAX_NOM_GRAPHE,
	       "Aucun graphe s"e_AIGU"lectionn"e_AIGU, TXT_COPIER);
#endif
#ifdef _ENGLISH_
    met_chaine(NORM_EXAFS, MU0_A_UTILISER, LG_MAX_NOM_GRAPHE,
	       "No selected graph", TXT_COPIER);
#endif
    }
    break;

  case MU0_THEORIQUE:
    OBJ_choisit(NORM_EXAFS, MU0_THEORIQUE , FORCE_OUI);
    OBJ_active (NORM_EXAFS, NORM_PROPOSER , methode != MU0_CONSTANT);
    OBJ_active (NORM_EXAFS, MU0_A_UTILISER, FORCE_NON);
    OBJ_active (NORM_EXAFS, MU0_CONSTANT  , FORCE_OUI);
    OBJ_active (NORM_EXAFS, MU0_HEITLER   , FORCE_OUI);
    OBJ_active (NORM_EXAFS, MU0_LANSBERGER, FORCE_OUI);
    OBJ_active (NORM_EXAFS, MU0_MAC_KALE  , FORCE_OUI);
    OBJ_active (NORM_EXAFS, MU0_VICTOREEN , FORCE_OUI);
    OBJ_active (NORM_EXAFS, NORM_CONSTANTE_1, FORCE_OUI);
    gere_normalise(methode, 0, 0);
    break;

  case MU0_CONSTANT:
    methode = objet;
    OBJ_choisit(NORM_EXAFS, objet, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_1, 2, " C", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_2, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_3, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_4, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_5, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_6, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_7, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_8, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_PROPOSER, FORCE_NON);
    break;

  case MU0_LANSBERGER:
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_1, 2, "E0", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_2, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_3, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_4, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_5, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_6, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_7, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_8, FORCE_NON);

  case MU0_HEITLER:
    methode = objet;
    OBJ_choisit(NORM_EXAFS, objet, FORCE_OUI);
    OBJ_active (NORM_EXAFS, NORM_PROPOSER, FORCE_OUI);
    break;

  case MU0_VICTOREEN:
    methode = objet;
    OBJ_choisit(NORM_EXAFS, objet, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_1, 2, "E0", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_PROPOSER, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_2, 2, "C0", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_2, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_3, 2, "D0", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_3, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_4, 2, "C1", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_4, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_5, 2, "D1", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_5, FORCE_OUI);
    OBJ_active(NORM_EXAFS, NORM_CTE_6, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_7, FORCE_NON);
    OBJ_active(NORM_EXAFS, NORM_CTE_8, FORCE_NON);
    break;

  case MU0_MAC_KALE: /* En ralit, McMaster... (McKale = phases et amplitudes) */
    methode = objet;
    OBJ_choisit(NORM_EXAFS, objet, FORCE_OUI);
/*    OBJ_active (NORM_EXAFS, NORM_PROPOSER, FORCE_OUI); Pas encore de tables intgres dans le programme */
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_1, 2, "E0", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_PROPOSER, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_2, 2, "A0", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_2, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_3, 2, "A1", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_3, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_4, 2, "A2", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_4, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_5, 2, "A3", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_5, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_6, 2, "B0", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_6, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_7, 2, "B1", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_7, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_8, 2, "B2", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_8, FORCE_OUI);
    met_chaine(NORM_EXAFS, NORM_TXT_CTE_9, 2, "B3", TXT_COPIER);
    OBJ_active(NORM_EXAFS, NORM_CTE_9, FORCE_OUI);
    break;

  case NORM_PROPOSER: /* Propositions de constantes en fonction de Z et de la mthode */
    propose_valeurs();
    break;

  default:
    OBJ_choisit(NORM_EXAFS, objet, ECHANGE);
  }
  return VRAI;
}

/* -------- Proposition de valeurs de constantes pour les calculs thoriques ------ */

static void propose_valeurs(void)
{
  extern float Seuil_K[], Seuil_L1[], Seuil_L2[], Seuil_L3[]; /* Les nergies de seuil - dans Theorie/absorption.c */
  double e_0;
  int element;

  element = prend_valeur(NORM_EXAFS, NORM_ELEMENT);
  if (element == 0) /* On essaye avec le symbole */
    element = ATM_num_atomique( lit_chaine(NORM_EXAFS, NORM_ELEMENT, 0, NULL, TXT_REMPLACE) );

  if ( (element < 1) || (element > 119) )
  {
    ATM_ERREUR(ELEMENT_INCONNU, element);
    return;
  }

  /* Valeur de E0 ncessaire pour toutes les mthodes, du fait du travail en chelle des K... */

  if ( OBJ_choisit(NORM_EXAFS, NORM_K, QUESTION) ) /* Seuil K */
    e_0 = (double) Seuil_K[element - 1];
  else if ( OBJ_choisit(NORM_EXAFS, NORM_L1, QUESTION) ) /* Seuil L1 */
    e_0 = (double) Seuil_L1[element - 1];
  else if ( OBJ_choisit(NORM_EXAFS, NORM_L2, QUESTION) ) /* Seuil L2 */
    e_0 = (double) Seuil_L2[element - 1];
  else if ( OBJ_choisit(NORM_EXAFS, NORM_L3, QUESTION) ) /* Seuil L3 */
    e_0 = (double) Seuil_L3[element - 1];
  else
    e_0 = -1.0;
    
  if (e_0 == -1.0) /* Valeur inconnue */
  {
  }
  else if (e_0 == -2.0) /* Valeur n'a pas de sens */
  {
  }

  met_fvaleur(NORM_EXAFS, NORM_CONSTANTE_1, e_0);

  if (methode == MU0_LANSBERGER) /* Aucune autre valeur  proposer */
    return;

  if (methode == MU0_VICTOREEN)
  {
    extern float CVict_K [], DVict_K [], /* Coefficients de Victoreen avant le seuil K  - dans Theorie/absorption.c */
                 CVict_L1[], DVict_L1[], /* Coefficients de Victoreen avant le seuil L1 */
                 CVict_L2[], DVict_L2[], /* Coefficients de Victoreen avant le seuil L2 */
                 CVict_L3[], DVict_L3[], /* Coefficients de Victoreen avant le seuil L3 */
                 CVict_M1[], DVict_M1[]; /* Coefficients de Victoreen avant le seuil M1 */
    double c1, d1, c2, d2;

    if ( OBJ_choisit(NORM_EXAFS, NORM_K, QUESTION) ) /* Seuil K */
    {
      c2 = (double) CVict_K[element - 1]; /* Aprs le seuil K : E > E(seuil K) */ 
      d2 = (double) DVict_K[element - 1];
      c1 = (double) CVict_L1[element - 1]; /* Avant le seuil K : E(K) > E > E(L1) */
      d1 = (double) DVict_L1[element - 1];
    }
    else if ( OBJ_choisit(NORM_EXAFS, NORM_L1, QUESTION) ) /* Seuil L1 */
    {
      c2 = (double) CVict_L1[element - 1];
      d2 = (double) DVict_L1[element - 1];
      c1 = (double) CVict_L2[element - 1];
      d1 = (double) DVict_L2[element - 1];
    }
    else if ( OBJ_choisit(NORM_EXAFS, NORM_L2, QUESTION) ) /* Seuil L2 */
    {
      c2 = (double) CVict_L2[element - 1];
      d2 = (double) DVict_L2[element - 1];
      c1 = (double) CVict_L3[element - 1];
      d1 = (double) DVict_L3[element - 1];
    }
    else if ( OBJ_choisit(NORM_EXAFS, NORM_L3, QUESTION) ) /* Seuil L3 */
    {
      c2 = (double) CVict_L3[element - 1];
      d2 = (double) DVict_L3[element - 1];
      c1 = (double) CVict_M1[element - 1];
      d1 = (double) DVict_M1[element - 1];
    }
    else if ( OBJ_choisit(NORM_EXAFS, NORM_M1, QUESTION) ) /* Seuil M1 */
    {
      c2 = (double) CVict_M1[element - 1];
      d2 = (double) DVict_M1[element - 1];
      c1 = -1; d1 = -1;
    }
    else
    {
      c1 = -1; d1 = -1;
      c2 = -1; d2 = -1;
    }

    if ( c1 == -1 )      /* Valeur inconnue */
    {
    }
    else if ( c1 == -2 ) /* Ne s'applique pas */
    {
    }
    else
      met_fvaleur(NORM_EXAFS, NORM_CTE_2, c1);

    if ( d1 == -1 )      /* Valeur inconnue */
    {
    }
    else if ( d1 == -2 ) /* Ne s'applique pas */
    {
    }
    else
      met_fvaleur(NORM_EXAFS, NORM_CTE_3, d1);

    if ( c2 == -1 )      /* Valeur inconnue */
    {
    }
    else if ( c2 == -2 ) /* Ne s'applique pas */
    {
    }
    else
      met_fvaleur(NORM_EXAFS, NORM_CTE_4, c2);

    if ( d2 == -1 )      /* Valeur inconnue */
    {
    }
    else if ( d2 == -2 ) /* Ne s'applique pas */
    {
    }
    else
      met_fvaleur(NORM_EXAFS, NORM_CTE_5, d2);

    return;
  }
}

/* -------- Prparation du graphe qui contiendra l'exafs normalis --------\
 |   Renvoi l'adresse du graphe cible (NULL si erreur survenue             |
 |                                                                         |
 | source  : le graphe que l'on veut normaliser                            |
 | nouveau : VRAI s'il faut crer un nouveau graphe                        |
 \------------------------------------------------------------------------ */

static Graphe *prepare_normalise(Graphe *source, boolean nouveau)
{
  Graphe *graphe;
  char titre[100];

  if (source == NULL) return NULL;

#ifdef _FRANCAIS_
      sprintf(titre, "%s (normalis"e_AIGU")", source->nom);
#endif
#ifdef _ENGLISH_
      sprintf(titre, "%s (normalised)", source->nom);
#endif

  if (nouveau == FAUX) /* On ne veut pas un nouveau graphe */
  {
    GRF_titre(source, titre);
    return source;
  }

  /* On veut un nouveau graphe */

  graphe = GRF_creer(source->nombre_point, source->type, GR_NORMAL,
		     titre, source->axe_x, source->axe_y, source->unit_x, source->unit_y);
  if (graphe == NULL)
    return NULL;

  /* On recopie les X dans le nouveau graphe */

  memcpy(graphe->x, source->x, source->nombre_point * sizeof(double) );
  graphe->x_min = source->x_min;
  graphe->x_max = source->x_max;

  return graphe;
}

/* -------- Calcul de mu suivant la formule de McMaster ----------------\
 |  Renvoi le coefficient d'absorption calcul                          |
 |                                                                      |
 | e : l'nergie pour laquelle on veut calculer                         |
 | a0, a1, a2, a3 : coefficients de McMaster                            |
 \--------------------------------------------------------------------- */

static double mu0_mcmaster(double e, double a0, double a1, double a2, double a3)
{
  double mu = a0;
  short int i;
  double ln_e = log(e);

  if (e <= 0) /* nergie incorrecte... risque de faire planter le log */
  {
    return 0;
  }

  mu += a1 * ln_e; /* mu = a0 + a1 * ln(E) */
  ln_e *= ln_e;
  mu += a2 * ln_e; /* mu = a0 + a1 * ln(E) + a2 * ln(E) */
  ln_e *= log(e);
  mu += a3 * ln_e; /* mu = a0 + a1 * ln(E) + a2 * ln(E) + a3 * ln(E) */

  return exp(mu); /* mu = exp( a0 + a1 * ln(E) + a2 * ln(E) + a3 * ln(E) ) */
}
