/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998, Emmanuel CURIS - Cration de ce fichier: 23 juin 1998

   Module de filtrage des spectres (essentiellement par transforme de Fourier)

   Paramtrage de la transforme de Fourier directe
   Source original rcupr de ../FOURIER.c et ../EXF_GEM.c
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "SYSTEME.H"

#include "EXAFS.H"
#include "ERREURS.H"
#include "GRAPHES.H"
#include "FOURIER.H"

/* ******** Variables locales ************************ */

#ifdef _FRANCAIS_
const char *Types_fenetres[] = 
{
  "Aucune",
  "Rectangulaire",
  "de Kaiser-Bessel",
  "Gaussienne",
  "de Hanning",
  "de Hamming",
  "F3 de Norton-Beer",
  "de Bartlett",
  "de Blackman"
};
#endif

#ifdef _ENGLISH_
const char *Types_fenetres[] = 
{
  "None",
  "Rectangular",
  "Kaiser-Bessel",
  "Gaussian",
  "Hanning",
  "Hamming",
  "F3 of Norton-Beer",
  "Bartlett",
  "Blackman"
};
#endif

static int num_tf_fenetre  = FEN_BESSEL;  /* Fentre  l'ouverture du module */

/* ******** Fonctions locales (pour la gestion de l'interface) ******** */

static void    RDS_tf(Fenetre *, int, int, int, int);
static boolean MSG_tf(Fenetre *fen);
static boolean GER_tf(int num_objet,int x, int y);
static boolean CLC_tf(Fenetre *fen, int bouton, int n_clic, int etat, int x, int y);

static void met_les_k(TF_infos *infos);

/* ******** Ralisation effective de l'interface ********************** */

/* -------- Appel de la fentre de transforme de Fourier ------------- */

void TF_ouvre(Graphe *graphe)
{
  Fenetre *fenetre;
  TF_infos *infos_tf;
  
  if (graphe == NULL) return; /* Rien  faire l... */
  if (graphe->drapeaux.protege == VRAI) /* Graphe utilis autre part */
  {
    EXF_ERREUR(GRF_PROTEGE);
    return;
  }

  fenetre = creer_mixte_fenetre(NAME|CLOSER|FULLER|INFO|SIZER|MOVER, PARAM_FOURIER, HAUT,
	                        RDS_tf, ( boolean (*)(Fenetre *, unsigned int, int) ) NULL,
				CLC_tf, MSG_tf, RCT_pointeur,
				( boolean (*)(Fenetre *) ) NULL);
  if (fenetre == NULL) return; /* Rien  faire... */
  (fenetre->arbre)->fonction.arbre = GER_tf;

  fenetre->drapeaux.redessin = VRAI;
  fenetre->drapeaux.pointeur = VRAI;
  fenetre->msg_masque.fermeture = VRAI; /* Messages traits par MSG_tf */

  infos_tf = (TF_infos *) fenetre->adresse;
  if (infos_tf) /* Dj des infos sur une TF -> on les perd */
  {
    if (infos_tf->graphe == graphe) /* Mme graphe -> pas  se fouler */
      return;
    else /* On efface ces infos pour en crer de nouvelles */
    {
      num_tf_fenetre = infos_tf->type_fenetre;
      GRF_efface(infos_tf->fenetre, VRAI, VRAI); /* On force l'effacement */
      infos_tf->graphe->drapeaux.protege = VRAI; /* Pas utilis -> on peut l'effacer */
      free(infos_tf->amplitude->graphes);
      free(infos_tf->amplitude);
      free(infos_tf->phase->graphes);
      free(infos_tf->phase);
      free(infos_tf);
    }
  }

  infos_tf = (TF_infos *) malloc( sizeof(TF_infos) ); /* On rserve les infos de TF */
  if (infos_tf == NULL)
  {
    ferme_fenetre(fenetre); /* On ferme la fentre : sert  rien... */
    ERREUR(MANQUE_MEMOIRE);
    return;
  }

  fenetre->adresse  = infos_tf;
  fenetre->pointeur = graphe;

#ifdef _FRANCAIS_
  infos_tf->fenetre = GRF_creer(graphe->nombre_point, GR_FENETRE, GR_IM_MOD,
				"Fen"e_CIRC"tre pour le calcul de la TF",
				graphe->axe_x, "Prise en compte", graphe->unit_x, UNIT_ARB);
#endif
#ifdef _ENGLISH_
  infos_tf->fenetre = GRF_creer(graphe->nombre_point, GR_FENETRE, GR_IM_MOD,
				"Window used for the FT calculation",
				graphe->axe_x, "Weight in the FT", graphe->unit_x, UNIT_ARB);
#endif

  if (infos_tf->fenetre == NULL) /* Mmoire insuffisante */
  {
    free(infos_tf);
    ferme_fenetre(fenetre);
    return;
  }
  infos_tf->fenetre->couleur = BLEU - 1;   /* Pour que le module apparaisse en bleu...    */
  infos_tf->fenetre->type    = GR_INTERNE; /* Graphe interne, ineffaable et tout et tout */
  infos_tf->fenetre->drapeaux.protege = VRAI; /* Protg -> ineffaable                   */

  infos_tf->graphe = graphe; /* Le graphe sur lequel on travaille */
  graphe->drapeaux.protege = VRAI;  /* Il est donc ineffaable... */

  /* Paramtres de la TF (fentre en particulier) */

  infos_tf->degre = prend_valeur(PARAM_FOURIER,FOUR_DEGRE);
  infos_tf->symetrique = OBJ_choisit(PARAM_FOURIER, TF_SYMETRIQUE, QUESTION); /* VRAI si fentre symtrique */

  infos_tf->k_min = graphe->x_min + 1; /* Intervalle pour lequel on calculera */
  infos_tf->k_max = graphe->x_max - 1;
  infos_tf->k_1 = infos_tf->k_min + 1;
  infos_tf->k_2 = infos_tf->k_max - 1;
  met_fvaleur(PARAM_FOURIER, FOUR_KMIN, infos_tf->k_min);
  met_fvaleur(PARAM_FOURIER, PARAM_TF_K1, infos_tf->k_1);
  met_fvaleur(PARAM_FOURIER, PARAM_TF_K2, infos_tf->k_2);
  met_fvaleur(PARAM_FOURIER, FOUR_KMAX, infos_tf->k_max);

  /* Calcul de la TF : fentre */

  infos_tf->type_fenetre = num_tf_fenetre; /* Fentre de Kaiser-Bessel par dfaut */
  infos_tf->num_depart   = FEN_RECTANGLE;  /* Affichage  partir de la premire */
  PDFL_textes(PARAM_FOURIER, FOUR_FEN_1, FOUR_FEN_4,
	      FOUR_FEN_AVANT, FOUR_FEN_APRES, FOUR_FEN_FOND, FOUR_FEN_BOUTON,
	      NOMBRE_FENETRES, (char **) Types_fenetres, NULL, num_tf_fenetre, 15, FEN_RECTANGLE);

  OBJ_active(PARAM_FOURIER, PARAM_TF_K1, Autorise_Symetrique(num_tf_fenetre) );
  OBJ_active(PARAM_FOURIER, PARAM_TF_K2, Autorise_Symetrique(num_tf_fenetre)
	                              && (infos_tf->symetrique) );
  OBJ_active(PARAM_FOURIER, TF_SYMETRIQUE, Autorise_Symetrique(num_tf_fenetre) );

  infos_tf->param_fenetre = prend_fvaleur(PARAM_FOURIER, FOUR_PARAM);
  TF_fenetre(infos_tf->fenetre, graphe, infos_tf->type_fenetre,
	     infos_tf->k_min, infos_tf->k_max, infos_tf->k_1, infos_tf->k_2,
	     infos_tf->param_fenetre, infos_tf->symetrique); /* On calcule la fentre */

  infos_tf->amplitude = NULL;
  infos_tf->phase = NULL;
}

/* ----------- Redessin de cette fentre ------------------------------------------------ */

static void RDS_tf(Fenetre *fenetre, int x, int y, int larg, int haut)
{
  Graphe *graphe;
  DES_graphe param_dessin = {NULL, NULL, 1, 0};
  TF_infos *infos;
  double x_echelle,y_echelle,x_gr,y_gr;
  int y_trace, haut_trace; /* Zone de dessin du graphe */

  infos = (TF_infos *) fenetre->adresse;

  vswr_mode(_PRG_station->handle,MD_REMPLACE);

  /* Fond blanc */

  vsf_color(_PRG_station->handle, BLANC);
  points[0] = x;        points[1] = y;
  points[2] = x + larg; points[3] = y + haut;
  v_bar(_PRG_station->handle,points);

  graphe = infos->graphe;

  /* Zone de dessin du graphe */

  y_trace    = fenetre->w_y    + OB_HEIGHT(arbres[PARAM_FOURIER].adresse,ROOT);
  haut_trace = fenetre->w_haut - OB_HEIGHT(arbres[PARAM_FOURIER].adresse,ROOT);

  /* Calcul des facteurs d'chelle */

  x_echelle = (double) (fenetre->w_larg) / (graphe->x_max - graphe->x_min);
  y_echelle = (double) (haut_trace) / (graphe->y_max - graphe->y_min);	

  /* Le graphe dont on fait la TF  */

  param_dessin.graphe  = graphe;
  param_dessin.couleur = NOIR; /* Graphe normal en noir */
  param_dessin.style   = 1;     /* Ligne continue        */
  TRC_graphe(_PRG_station, &param_dessin, fenetre->w_x, y_trace,
	     graphe->x_min, graphe->x_max, graphe->y_min, graphe->y_max,
	     x_echelle, y_echelle);

  /* Dessin de la fentre */

  param_dessin.graphe  = infos->fenetre;
  param_dessin.couleur = BLEU - 1;  /* Fentre en bleu, si possible */
  TRC_graphe(_PRG_station, &param_dessin, fenetre->w_x, y_trace,
             infos->fenetre->x_min, infos->fenetre->x_max,
	     infos->fenetre->y_min, infos->fenetre->y_max,
             x_echelle, haut_trace);

  /* Dessin du graphe multipli par la fentre  - en violet */
  vsl_color(_PRG_station->handle, VIOLET);
  vsl_type (_PRG_station->handle, 1);

  if ( graphe->drapeaux.complexe == GR_NORMAL ) /* Graphe normal */
  {
    unsigned long i;
    unsigned int compteur = 0;
    register double   *zone_x;
    register Complexe *zone_y;

    zone_x = infos->fenetre->x; zone_y = (Complexe *) (infos->fenetre->y);
    for (i = 0 ; i < infos->fenetre->nombre_point ; i++)
    {
      x_gr = *zone_x; y_gr = zone_y->re; /* Partie relle = fentre, partie imaginaire nulle */

      if (infos->degre > 0)
      {
	int i;
	for (i = 0 ; i <= infos->degre ; i++)
	  y_gr *= x_gr;
      }
      y_gr *= *(graphe->y + i);

      points[compteur]   = fenetre->w_x + x_echelle * (x_gr - graphe->x_min);
      points[compteur+1] = y_trace      + y_echelle * (graphe->y_max - y_gr);
      compteur += 2;
      zone_x++;
      zone_y++;
      if (compteur >= 256) /* Tableau de points complet */
      {
	v_pline(_PRG_station->handle,128,points);
	points[0] = points[254];
	points[1] = points[255];
	compteur = 2;
      }
    }
    v_pline(_PRG_station->handle,(int) (compteur-1)/2,points);	
  }

  /* Mise en place des barres pour k_min (rouge) et k_max (vert) */

  vsl_color(_PRG_station->handle, ROUGE);
  points[0] = fenetre->w_x + x_echelle * (infos->k_min - graphe->x_min);
  points[1] = y_trace;
  points[2] = points[0];
  points[3] = y_trace + haut_trace;
  v_pline(_PRG_station->handle,2,points);
	
  vsl_color(_PRG_station->handle, VERT);
  points[0] = fenetre->w_x + x_echelle * (infos->k_max - graphe->x_min);
  points[2] = points[0];
  v_pline(_PRG_station->handle,2,points);

  /* Mise en place des barres pour k_1 et k_2, si la fentre le ncessite */

  if ( Autorise_Symetrique(infos->type_fenetre) )
  {
    double k_2;

    vsl_color(_PRG_station->handle, ROUGE_FONCE);
    points[0] = fenetre->w_x + x_echelle * (infos->k_1 - graphe->x_min);
    points[2] = points[0];
    v_pline(_PRG_station->handle, 2, points);	
    
    if (infos->symetrique)
      k_2 = infos->k_max - (infos->k_1 - infos->k_min);
    else
      k_2 = infos->k_2;

    vsl_color(_PRG_station->handle, VERT_FONCE);
    points[0] = fenetre->w_x + x_echelle * (k_2 - graphe->x_min);
    points[2] = points[0];
    v_pline(_PRG_station->handle, 2, points);	
  }
}

/* -------- Traitement des messages reus par la fentre --------------------------- */

boolean MSG_tf(Fenetre *fenetre)
{
  Graphe *graphe, *tf_fenetre, *fourier, *inter;
  TF_infos *infos; /* Les infos pour la fentre, pour la TF */
  int reponse;

  infos      = (TF_infos *) fenetre->adresse;
  graphe     = infos->graphe;
  tf_fenetre = infos->fenetre;

  if ( buffer[0] == WM_CLOSED ) /* On demande la fermeture */
  {
    reponse = form_alert(1,"[2][ Quitter le param"e_AIGU"trage de la fen"e_CIRC"tre | Que faire pour le calcul de la TF ? ][ Calculer | Rien | Annuler ]");
    if (reponse == 3) return VRAI; /* On annule */

    num_tf_fenetre = infos->type_fenetre;
    if (reponse == 2) /* On quitte sans calculer la TF */
    {
      GRF_efface(tf_fenetre, VRAI, VRAI); /* On efface ce graphe          */
      graphe->drapeaux.protege = FAUX;
      free(infos->amplitude);
      free(infos->phase);
      free(infos);                /* On libre la mmoire         */
      ferme_fenetre(fenetre);     /* On ferme la fentre          */
      return VRAI;
    }

    /* On ferme la fentre : les choses srieuses commencent... */

    FEN_ferme(fenetre, FAUX); /* vite un sale message de redessin */

    /* Prparation de la TF : fourier = TF calcul; tf_fenetre = infos->fenetre = source */

    fourier = TF_init(infos); 
    if (fourier == NULL) /* Calcul annul (utilisateur ou problme) */
    {
      FEN_reouvre(fenetre);
      return VRAI;
    }

    ferme_fenetre(fenetre); /* On ferme entirement la fentre */

    /* On fait ventuellement les corrections de phase et d'amplitude */

    if ( OBJ_choisit(DEBUT_TF, TF_COR_AMP, QUESTION) == VRAI ) /* Correction d'amplitude */
    {
      inter = ( (infos->amplitude)->graphes[infos->amplitude->num_select] );
      TF_coramp(tf_fenetre,inter); /* Ralisation de la correction d'amplitude */
    }

    if ( OBJ_choisit(DEBUT_TF, TF_COR_PHASE, QUESTION) == VRAI ) /* Correction de phase */
    {
      inter = ( (infos->phase)->graphes[infos->phase->num_select] );

      if ( TF_corpha(tf_fenetre,inter,VRAI) == FAUX ) /* Ralisation de la correction de phase */
      {
	OBJ_choisit(DEBUT_TF, TF_COR_PHASE, FORCE_NON);
      }
    }

    /* Calcul de la transforme de Fourier - le spectre  TFer est dans tf_fenetre */

    if ( OBJ_choisit(DEBUT_TF, TF_RAPIDE, QUESTION) == VRAI ) /* On veut la mthode rapide */
    {
      double r_max;

      TF_rapide(tf_fenetre, fourier);

      r_max = prend_fvaleur(DEBUT_TF, R_MAX); /* Valeur de Rmax demande par l'utilisateur */
      if (r_max < fourier->x_max) /* On ralise la troncature */
      {
	double *x;
	unsigned long i;

	for (x = fourier->x , i = 0 ; *x <= r_max ; x++, i++);
	fourier->nombre_point = i; /* Troncature _apparente_, la fin est toujours en mmoire... */
      }
    }
    else
    {
      TF_simple(tf_fenetre, fourier);
    }

    GRF_bornes(fourier, ORDONNE, DESORDONNE); /* On rcupre les bornes correctes */
		
    sprintf(_intermediaire, "T.F. de %s", graphe->nom);
    GRF_titre(fourier, _intermediaire); /* On met le titre */

    /* On peut supposer que l'on a fini ici... */
    GRF_efface(tf_fenetre, VRAI, VRAI);
    fourier->drapeaux.protege = FAUX; /* On a fini de calculer sur ces graphes */
    graphe->drapeaux.protege  = FAUX;

    free(infos->phase);     infos->phase     = NULL;
    free(infos->amplitude); infos->amplitude = NULL;
    fourier->pointeur = infos; /* On conserve les paramtres de calcul de la TF... */

    GRF_aller(fourier); /* On met ce graphe en graphe courant */
    
    if (TT_exafs == VRAI) /* On veut faire tout l'E.X.A.F.S. */
    {
      TFI_ouvre(fourier); /* On appelle pour la TF inverse */

      TT_exafs = FAUX;
    }

    return VRAI;
  }

  return VRAI; /* On a trait tout ce que l'on voulait traiter */
}

/* --------- Gestion d'un clic sur la fentre de TF --------------------------------------- */

boolean CLC_tf(Fenetre *fenetre, int bouton, int n_clic, int etat, int x, int y)
{
  TF_infos *infos;
  double x_echelle,x_clic;

  if (y < fenetre->w_y + OB_HEIGHT(arbres[PARAM_FOURIER].adresse, ROOT) )
    return VRAI;

  infos = (TF_infos *) fenetre->adresse;

  x_echelle = ( (double) fenetre->w_larg ) / (infos->graphe->x_max - infos->graphe->x_min);
  x_clic  = x - fenetre->w_x;
  x_clic /= x_echelle;
  x_clic += infos->graphe->x_min;

  if (bouton == 1) /* Bouton gauche */
  {
    if ( (etat & K_LSHIFT) || (etat & K_RSHIFT) ) /* On appuie sur shift en mme temps */
    {
      if ( Autorise_Symetrique( infos->type_fenetre ) )
	infos->k_1 = x_clic;
      else
	infos->k_min = x_clic;
    }
    else
      infos->k_min = x_clic;
  }
  else /* Bouton droit */
  {
    if ( (etat & K_LSHIFT) || (etat & K_RSHIFT) ) /* On appuie sur shift en mme temps */
    {
      if ( Autorise_Symetrique( infos->type_fenetre ) && (infos->symetrique == FAUX) )
	infos->k_2 = x_clic;
      else
	infos->k_max = x_clic;
    }
    else
      infos->k_max = x_clic;
  }

  met_les_k(infos);

  TF_fenetre(infos->fenetre, infos->graphe, infos->type_fenetre, infos->k_min, infos->k_max, 
	     infos->k_1, infos->k_2, infos->param_fenetre, infos->symetrique);

  FEN_redessin(fenetre); /* On redessine : fentre a chang... */
  return VRAI;
}

/* -------- Prparation de la TF (choix de la fentre principalement) ------------------------- */

static boolean GER_tf(int num_objet, int x_souris, int y_souris)
{
  Fenetre *fen_cour;
  Graphe *fenetre;
  OBJECT *adresse = arbres[PARAM_FOURIER].adresse;
  TF_infos *infos;
  extern const char *Types_fenetres[];

  fen_cour = ARB_fenetre(arbres + PARAM_FOURIER); /* Fentre actuelle */
  infos = (TF_infos *) fen_cour->adresse;
  fenetre = infos->fenetre;

  if ( (num_objet >= FOUR_FEN_1) && (num_objet <= FOUR_FEN_4) )
  {
    unsigned long nv_fenetre = infos->type_fenetre; /* Nouvelle fentre choisie */
    unsigned long prem_visible = infos->num_depart;

    GDFL_textes(PARAM_FOURIER, FOUR_FEN_1, FOUR_FEN_4,
		FOUR_FEN_AVANT, FOUR_FEN_APRES, FOUR_FEN_FOND, FOUR_FEN_BOUTON,
		NOMBRE_FENETRES, (char **) Types_fenetres, NULL, &nv_fenetre, 15,
		&prem_visible, num_objet, y_souris);

    if ( nv_fenetre != infos->type_fenetre )
    {
      boolean etat;

      etat = (nv_fenetre == FEN_HANNING) || (nv_fenetre == FEN_HAMMING)
	  || (nv_fenetre == FEN_BARTLETT); /* VRAI -> asymtrie possible */
      OBJ_active(PARAM_FOURIER, TF_SYMETRIQUE, etat);
      OBJ_active(PARAM_FOURIER, PARAM_TF_K1, etat);

      etat &= ! (OBJ_choisit(PARAM_FOURIER, TF_SYMETRIQUE, QUESTION));
      OBJ_active(PARAM_FOURIER, PARAM_TF_K2, etat);

      etat = (nv_fenetre == FEN_HAMMING) || (nv_fenetre == FEN_NORTON_F3)
	  || (nv_fenetre == FEN_BESSEL)  || (nv_fenetre == FEN_GAUSSIENNE);
      OBJ_active(PARAM_FOURIER, FOUR_PARAM, etat); /* Seule fentre avec un paramtre... */

      infos->type_fenetre = (int) nv_fenetre; /* On choisit la nouvelle fentre */
      TF_fenetre(fenetre, infos->graphe, infos->type_fenetre,
		 infos->k_min, infos->k_max, infos->k_1, infos->k_2, 
		 infos->param_fenetre, infos->symetrique); /* On calcule la fentre */
      FEN_redessin(fen_cour);
    }
  }

  switch (num_objet)
  {
  case FOUR_FEN_AVANT:      /* Dfilement du slecteur de fentres */
  case FOUR_FEN_APRES:
  case FOUR_FEN_BOUTON - 1:
  case FOUR_FEN_BOUTON:
    {
      unsigned long nv_fenetre = infos->type_fenetre,
	prem_visible = infos->num_depart;
      GDFL_textes(PARAM_FOURIER, FOUR_FEN_1, FOUR_FEN_4,
		  FOUR_FEN_AVANT, FOUR_FEN_APRES, FOUR_FEN_FOND, FOUR_FEN_BOUTON,
		  NOMBRE_FENETRES, (char **) Types_fenetres, NULL, &nv_fenetre, 15,
		  &prem_visible, num_objet, y_souris);
      infos->num_depart = (int) prem_visible; /* Il n'y aura jamais max_int fentres... */
    }
    break;

  case FOUR_KMIN: /* Demande de k min  l'utilisateur */
    met_chaine(CHOIX_VALEUR, CHXVAL_UNITE, 0, ANGSTROM"^-1", TXT_REMPLACE);
#ifdef _FRANCAIS_
    met_chaine(CHOIX_VALEUR, CHXVAL_TITRE, 0, "D"e_AIGU"but de la fen"e_CIRC"tre", TXT_REMPLACE);
#endif
#ifdef _ENGLISH_
    met_chaine(CHOIX_VALEUR, CHXVAL_TITRE, 0, "Beginning of the window", TXT_REMPLACE);
#endif
    met_fvaleur(CHOIX_VALEUR, CHXVAL_VALEUR, infos->k_min);

    ext_formdo(CHOIX_VALEUR);

    infos->k_min = prend_fvaleur(CHOIX_VALEUR, CHXVAL_VALEUR);
    met_les_k(infos);

    TF_fenetre(fenetre, infos->graphe, infos->type_fenetre,
	       infos->k_min, infos->k_max, infos->k_1, infos->k_2, 
	       infos->param_fenetre, infos->symetrique); /* On calcule la fentre */
    FEN_redessin(fen_cour);
    return VRAI;

  case PARAM_TF_K1: /* Demande de k_1  l'utilisateur */
    met_chaine(CHOIX_VALEUR, CHXVAL_UNITE, 0, ANGSTROM"^-1", TXT_REMPLACE);
    met_chaine(CHOIX_VALEUR, CHXVAL_TITRE, 0, "Fin de mont"e_AIGU"e de la fen"e_CIRC"tre", TXT_REMPLACE);
    met_fvaleur(CHOIX_VALEUR, CHXVAL_VALEUR, infos->k_1);

    ext_formdo(CHOIX_VALEUR);

    infos->k_1 = prend_fvaleur(CHOIX_VALEUR, CHXVAL_VALEUR);
    met_les_k(infos);

    TF_fenetre(fenetre, infos->graphe, infos->type_fenetre,
	       infos->k_min, infos->k_max, infos->k_1, infos->k_2, 
	       infos->param_fenetre, infos->symetrique); /* On calcule la fentre */
    FEN_redessin(fen_cour);
    return VRAI;

  case PARAM_TF_K2: /* Demande de k_2  l'utilisateur */
    met_chaine(CHOIX_VALEUR, CHXVAL_UNITE, 0, ANGSTROM"^-1", TXT_REMPLACE);
    met_chaine(CHOIX_VALEUR, CHXVAL_TITRE, 0, "D"e_AIGU"but de descente de la fen"e_CIRC"tre", TXT_REMPLACE);
    met_fvaleur(CHOIX_VALEUR, CHXVAL_VALEUR, infos->k_2);

    ext_formdo(CHOIX_VALEUR);

    infos->k_2 = prend_fvaleur(CHOIX_VALEUR, CHXVAL_VALEUR);
    met_les_k(infos);

    TF_fenetre(fenetre, infos->graphe, infos->type_fenetre,
	       infos->k_min, infos->k_max, infos->k_1, infos->k_2, 
	       infos->param_fenetre, infos->symetrique); /* On calcule la fentre */
    FEN_redessin(fen_cour);
    return VRAI;

  case FOUR_KMAX: /* Demande de k max  l'utilisateur */
    met_chaine(CHOIX_VALEUR, CHXVAL_UNITE, 0, ANGSTROM"^-1", TXT_REMPLACE);
    met_chaine(CHOIX_VALEUR, CHXVAL_TITRE, 0, "Fin de la fen"e_CIRC"tre", TXT_REMPLACE);
    met_fvaleur(CHOIX_VALEUR, CHXVAL_VALEUR, infos->k_max);

    ext_formdo(CHOIX_VALEUR);

    infos->k_max = prend_fvaleur(CHOIX_VALEUR, CHXVAL_VALEUR);
    met_les_k(infos);

    TF_fenetre(fenetre, infos->graphe, infos->type_fenetre,
	       infos->k_min, infos->k_max, infos->k_1, infos->k_2, 
	       infos->param_fenetre, infos->symetrique); /* On calcule la fentre */
    FEN_redessin(fen_cour);
    break;

  case FOUR_DEGRE:
    met_chaine(CHOIX_VALEUR, CHXVAL_UNITE, 0, NULL, TXT_REMPLACE);
    met_chaine(CHOIX_VALEUR, CHXVAL_TITRE, 0, "Pond"e_AIGU"rer par k"n_EXPOSANT, TXT_REMPLACE);
    met_fvaleur(CHOIX_VALEUR, CHXVAL_VALEUR, (double) infos->degre);

    ext_formdo(CHOIX_VALEUR);
					
    infos->degre = (char) prend_fvaleur(CHOIX_VALEUR, CHXVAL_VALEUR);
    met_valeur(PARAM_FOURIER, FOUR_DEGRE, infos->degre);

    FEN_redessin(fen_cour);
    break;

  case FOUR_PARAM:
    met_chaine(CHOIX_VALEUR, CHXVAL_UNITE, 0, NULL, TXT_REMPLACE);
    met_chaine(CHOIX_VALEUR, CHXVAL_TITRE, 0, "Param"e_GRAVE"tre de fen"e_CIRC"tre", TXT_REMPLACE);
    met_fvaleur(CHOIX_VALEUR,CHXVAL_VALEUR,infos->param_fenetre);

    ext_formdo(CHOIX_VALEUR);
					
    infos->param_fenetre = prend_fvaleur(CHOIX_VALEUR,CHXVAL_VALEUR);
    sprintf((char *) OB_SPEC(adresse,FOUR_PARAM),"%6.3f%c",infos->param_fenetre,0);

    TF_fenetre(fenetre, infos->graphe, infos->type_fenetre,
	       infos->k_min, infos->k_max, infos->k_1, infos->k_2, 
	       infos->param_fenetre, infos->symetrique); /* On calcule la fentre */
    FEN_redessin(fen_cour);
    break;

  case TF_SYMETRIQUE:
    infos->symetrique = OBJ_choisit(PARAM_FOURIER, TF_SYMETRIQUE, ECHANGE);
    OBJ_active(PARAM_FOURIER, PARAM_TF_K2, !(infos->symetrique)); /* Aucun sens si fentre symtrique... */
    TF_fenetre(fenetre, infos->graphe, infos->type_fenetre,
	       infos->k_min, infos->k_max, infos->k_1, infos->k_2, 
	       infos->param_fenetre, infos->symetrique); /* On calcule la fentre */
    FEN_redessin(fen_cour);
  }

  return VRAI;
}

/* -------- Met les valeurs des k aprs tri (pour la cohrence) -------- */

static void met_les_k(TF_infos *infos)
{
  if (infos->k_min > infos->k_max) /* On vrifie la cohrence */
  {
    double k = infos->k_min;
    infos->k_min = infos->k_max;
    infos->k_max = k;
  }
  if ( Autorise_Symetrique( infos->type_fenetre ) )
  {
    if (infos->symetrique)
    {
      if ( infos->k_1 < infos->k_min )
      {
	double k = infos->k_min;
	infos->k_min = infos->k_1;
	infos->k_1 = k;
      }
      if ( infos->k_1 > infos->k_max )
      {
	double k = infos->k_max;
	infos->k_max = infos->k_1;
	infos->k_1 = k;
      }
    }
    else
    {
      double k[4];
      int i, j;

      k[0] = infos->k_min;
      k[1] = infos->k_1;
      k[2] = infos->k_2;
      k[3] = infos->k_max;

      for (i = 0 ; i < 4 ; i++)
	for (j = i ; j < 4 ; j++)
	  if ( k[i] > k[j] )
	  {
	    double l = k[i];
	    k[i] = k[j]; k[j] = l;
	  }

      infos->k_min = k[0];
      infos->k_1   = k[1];
      infos->k_2   = k[2];
      infos->k_max = k[3];
    }
  }

  met_fvaleur(PARAM_FOURIER, FOUR_KMIN, infos->k_min);
  if ( Autorise_Symetrique( infos->type_fenetre ) )
  {
    met_fvaleur(PARAM_FOURIER, PARAM_TF_K1, infos->k_1);
    if (infos->symetrique == FAUX)
      met_fvaleur(PARAM_FOURIER, PARAM_TF_K2, infos->k_2);
  }
  met_fvaleur(PARAM_FOURIER, FOUR_KMAX, infos->k_max);
}

/* -------- Paramtrages du calcul de la transforme de Fourier -------- */

/* ~~~~~~~~~~~~~~~~~ Calcul proprement dit de la Transforme de Fourier ~~~~~~~~~~~~~~~~~~ */

/* ----------------- Initialisation du graphe qui contiendra la TF ------------------------\
 |   Renvoie le graphe prpar pour le calcul de la TF
 |
 | infos_tf     : informations ncessaires au calcul de la transforme de Fourier          |
 | r_min, r_max : bornes de calcul de la transforme de Fourier
 | pas          : pas  utiliser pour ce calcul
 \---------------------------------------------------------------------------------------- */

Graphe *TF_init(TF_infos *infos_tf)
{
  OBJECT *adresse = arbres[DEBUT_TF].adresse;
  Graphe *fourier;
  int retour;

  arbres[DEBUT_TF].pointeur = infos_tf;     /* Infos complmentaires pour l'arbre... */

  infos_tf->amplitude = GRF_par_type(GR_EXAFSAMP, FAUX); /* Prparation de la liste des graphes d'amplitudes */
  OBJ_active(DEBUT_TF, TF_COR_AMP, infos_tf->amplitude != NULL);
  if (infos_tf->amplitude == NULL)                 /* Prparation n'a pu tre faite */
    OBJ_choisit(DEBUT_TF, TF_COR_AMP, FORCE_NON);

  if (OB_STATE(adresse,TF_COR_AMP) & SELECTED)
  {
    OB_FLAGS(adresse,TF_COR_AMP_1 - 1) &= ~HIDETREE;
    OB_FLAGS(adresse,TF_COR_AMP_AV)    &= ~HIDETREE;
    OB_FLAGS(adresse,TF_COR_AMP_FOND)  &= ~HIDETREE;
    OB_FLAGS(adresse,TF_COR_AMP_AP)    &= ~HIDETREE;
  }
  else
  {
    OB_FLAGS(adresse,TF_COR_AMP_1 - 1) |= HIDETREE;
    OB_FLAGS(adresse,TF_COR_AMP_AV)    |= HIDETREE;
    OB_FLAGS(adresse,TF_COR_AMP_FOND)  |= HIDETREE;
    OB_FLAGS(adresse,TF_COR_AMP_AP)    |= HIDETREE;
  }

  infos_tf->phase = GRF_par_type(GR_EXAFSPHA, FAUX);     /* Prparation de la liste des graphes de phase     */
  if (infos_tf->phase == NULL)
  {
    OB_STATE(adresse,TF_COR_PHASE) &= ~SELECTED;
    OB_STATE(adresse,TF_COR_PHASE) |=  DISABLED;  /* Correction de phase impossible... */
  }
  else
  {
    OB_STATE(adresse,TF_COR_PHASE) &= ~DISABLED;
  }
  if (OB_STATE(adresse,TF_COR_PHASE) & SELECTED)
  {
    OB_FLAGS(adresse,TF_COR_PHA_1 - 1) &= ~HIDETREE;
    OB_FLAGS(adresse,TF_COR_PHA_AV)    &= ~HIDETREE;
    OB_FLAGS(adresse,TF_COR_PHA_FOND)  &= ~HIDETREE;
    OB_FLAGS(adresse,TF_COR_PHA_AP)    &= ~HIDETREE;
  }
  else
  {
    OB_FLAGS(adresse,TF_COR_PHA_1 - 1) |= HIDETREE;
    OB_FLAGS(adresse,TF_COR_PHA_AV)    |= HIDETREE;
    OB_FLAGS(adresse,TF_COR_PHA_FOND)  |= HIDETREE;
    OB_FLAGS(adresse,TF_COR_PHA_AP)    |= HIDETREE;
  }

  GRF_affiche(infos_tf->amplitude, DEBUT_TF, TF_COR_AMP_1, TF_COR_AMP_4, 
	      TF_COR_AMP_AV, TF_COR_AMP_AP, TF_COR_AMP_FOND, TF_COR_AMP_BOUT);
  GRF_affiche(infos_tf->phase,     DEBUT_TF, TF_COR_PHA_1, TF_COR_PHA_4,
	      TF_COR_PHA_AV, TF_COR_PHA_AP, TF_COR_PHA_FOND, TF_COR_PHA_BOUT);

  retour = ext_formdo(DEBUT_TF); /* On paramtre le calcul de la TF (autre que la fentre) */
  if (retour == FOUR_ANNULE)     /* Finalement, on reste...                                */
  {
    free(infos_tf->phase);
    free(infos_tf->amplitude);
    return NULL;
  }

  /* On prpare tout pour la TF */

  if ( OBJ_choisit(DEBUT_TF, TF_RAPIDE, QUESTION) ) /* TF rapide */
  {
    fourier = TF_prep_rapide(infos_tf, prend_fvaleur(DEBUT_TF, R_MIN), prend_fvaleur(DEBUT_TF, R_MAX),
			     prend_fvaleur(DEBUT_TF,FOUR_PAS));
  }
  else
  {
    fourier = TF_prep_normal(infos_tf, prend_fvaleur(DEBUT_TF, R_MIN), prend_fvaleur(DEBUT_TF, R_MAX),
			     prend_fvaleur(DEBUT_TF,FOUR_PAS));
  }

  if (fourier == NULL) /* Une erreur s'est produite */
  {
    free(infos_tf->phase);     /* On libre : on n'a plus besoin de a... */
    free(infos_tf->amplitude);
    return NULL;
  }

  return fourier; /* Prparation termine */	
}

/* --------------- Lancement de la TF (choix des graphes de correction, entre autres ----------------------\
 |   Gestion de l'arbre GEM de paramtrage des paramtres de calcul autres que la fentre                  |
 \-------------------------------------------------------------------------------------------------------- */

boolean GER_debTF(int num_objet, int x, int y)
{
  TF_infos *infos = (TF_infos *) (arbres[DEBUT_TF].pointeur);
  OBJECT *adresse = arbres[DEBUT_TF].adresse;
  int nv_obj;
  boolean etat;

  if ( (num_objet >= TF_COR_AMP_1) && (num_objet <= TF_COR_AMP_AP) )
  {
    return GRF_selecteur(infos->amplitude, DEBUT_TF, TF_COR_AMP_1, TF_COR_AMP_4,
			 TF_COR_AMP_AV, TF_COR_AMP_AP, TF_COR_AMP_FOND, TF_COR_AMP_BOUT,
			 num_objet, x, y);
  }

  if ( (num_objet >= TF_COR_PHA_1) && (num_objet <= TF_COR_PHA_AP) )
  {
    return GRF_selecteur(infos->phase, DEBUT_TF, TF_COR_PHA_1, TF_COR_PHA_4,
			 TF_COR_PHA_AV, TF_COR_PHA_AP, TF_COR_PHA_FOND, TF_COR_PHA_BOUT,
			 num_objet, x, y);
  }

  switch (num_objet)
  {
  case TF_COR_AMP: /* Activation ou non de la correction d'amplitude */

    etat = OBJ_choisit(DEBUT_TF, TF_COR_AMP, ECHANGE);

    OBJ_cache(DEBUT_TF, TF_COR_AMP_1 - 1, !etat);
    OBJ_cache(DEBUT_TF, TF_COR_AMP_AV, !etat);
    OBJ_cache(DEBUT_TF, TF_COR_AMP_FOND, !etat);
    OBJ_cache(DEBUT_TF, TF_COR_AMP_AP, !etat);

#if _OBJ_cache_ne_fonctionne_pas

    if (OB_STATE(adresse,TF_COR_AMP) & SELECTED) /* On veut utiliser la correction d'amplitude */
    {
      OB_FLAGS(adresse,TF_COR_AMP_1 - 1) &= ~HIDETREE;
      OB_FLAGS(adresse,TF_COR_AMP_AV)    &= ~HIDETREE;
      OB_FLAGS(adresse,TF_COR_AMP_FOND)  &= ~HIDETREE;
      OB_FLAGS(adresse,TF_COR_AMP_AP)    &= ~HIDETREE;
    }
    else
    {
      OB_FLAGS(adresse,TF_COR_AMP_1 - 1) |= HIDETREE;
      OB_FLAGS(adresse,TF_COR_AMP_AV)    |= HIDETREE;
      OB_FLAGS(adresse,TF_COR_AMP_FOND)  |= HIDETREE;
      OB_FLAGS(adresse,TF_COR_AMP_AP)    |= HIDETREE;
    }
    
    dessine_objet(DEBUT_TF,TF_COR_AMP_AV);
    dessine_objet(DEBUT_TF,TF_COR_AMP_FOND);
    dessine_objet(DEBUT_TF,TF_COR_AMP_AP);
    dessine_objet(DEBUT_TF,TF_COR_AMP_1 - 1);

#endif

    return VRAI;

  case TF_COR_PHASE: /* Activation ou non de la correction de phase */
    form_button(adresse,num_objet,1,&nv_obj);
    if (OB_STATE(adresse,TF_COR_PHASE) & SELECTED)
    {
      OB_FLAGS(adresse,TF_COR_PHA_1 - 1) &= ~HIDETREE;
      OB_FLAGS(adresse,TF_COR_PHA_AV)    &= ~HIDETREE;
      OB_FLAGS(adresse,TF_COR_PHA_FOND)  &= ~HIDETREE;
      OB_FLAGS(adresse,TF_COR_PHA_AP)    &= ~HIDETREE;
    }
    else
    {
      OB_FLAGS(adresse,TF_COR_PHA_1 - 1) |= HIDETREE;
      OB_FLAGS(adresse,TF_COR_PHA_AV)    |= HIDETREE;
      OB_FLAGS(adresse,TF_COR_PHA_FOND)  |= HIDETREE;
      OB_FLAGS(adresse,TF_COR_PHA_AP)    |= HIDETREE;
    }
    dessine_objet(DEBUT_TF,TF_COR_PHA_AV);
    dessine_objet(DEBUT_TF,TF_COR_PHA_FOND);
    dessine_objet(DEBUT_TF,TF_COR_PHA_AP);
    dessine_objet(DEBUT_TF,TF_COR_PHA_1 - 1);
    return VRAI;

  case TF_RAPIDE:
    OBJ_choisit(DEBUT_TF, TF_RAPIDE, FORCE_OUI);
    OBJ_active(DEBUT_TF, R_MIN, FORCE_NON); /* Pas de choix pour Rmin... */
    break;

  case TF_NORMALE:
    OBJ_choisit(DEBUT_TF, TF_NORMALE, FORCE_OUI);
    OBJ_active(DEBUT_TF, R_MIN, FORCE_OUI); /* Choix de Rmin possible */
    break;

  default:
    form_button(adresse,num_objet,1,&nv_obj);
  }

  return VRAI; /* Trait */
}
