/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998, Emmanuel CURIS - Cration de ce fichier : 15 septembre 1998

   Module de filtrage des spectres (essentiellement par transforme de Fourier)

   Corrections de phase et d'amplitude (avant / aprs calcul de la TF / TFI)
   Source original rcupr de ../FOURIER.c et ../AJUSTE.c
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "SYSTEME.H"

#include "ERREURS.H"
#include "EXAFS.H"
#include "GRAPHES.H"
#include "FOURIER.H"
#include "LISSAGE.H"
#include "SCIENCE.H" /* Si jamais M_PI n'avait pas t dfini dans <math.h> */

/* ******** Variables locales ************************ */

#define PAS_DE_LPM 0

/* ******** Fonctions locales ************************ */

static boolean GER_amplitude(int objet, int x, int y);

/* ******** Ralisation pratique des fonctions ******* */

/* ~~~~~~~~ Correction en phase et amplitude d'une TF ~~~~~~~~ */

/* ------ Correction en amplitude du spectre dont on veut la TF ---------------------------------\
 |   Renvoie FAUX si il y a eu un problme lors de la correction                                 |
 |                                                                                               |
 | graphe  : le graphe dont on veut raliser la correction                                       |
 | correct : le graphe contenant les donnes d'amplitude servant  corriger                      |
 \---------------------------------------------------------------------------------------------- */

boolean TF_coramp(Graphe *graphe, Graphe *correct)
{
  double *b = NULL, *c = NULL, *d = NULL; /* Tableaux pour la spline */
  double amp;
  register double *zone_x;
  int retour;
  unsigned long i;

  if ( (correct == NULL) || (graphe == NULL) ) return FAUX;

  if ( (correct->x_min >= graphe->x_max ) || (correct->x_max <= graphe->x_min) )
  {
    form_alert(1,"[3][ Les intervalles en X | sont disjoints. | La correction est impossible ][ Annuler ]");
    return FAUX;
  }

  /* On calcule la spline qui permettra ensuite la correction */
  retour = SPL_tt_prep( correct, &b, &c, &d);
  if (retour == -1)
  {
    return FAUX; /* Problme pour le calcul de la SPLINE */
  }

  zone_x = graphe->x;
  if (graphe->drapeaux.complexe)
  {
    register Complexe *zone_y = (Complexe *) graphe->y;

    for (i = 0 ; i < graphe->nombre_point ; i++)
    {
      amp = SPL_tt_calc( correct, *zone_x, b, c, d);
      if (amp) /* Pas de risque de division par zro */
      {
	zone_y->re /= amp;
	zone_y->im /= amp;
      }
      else
      {
	EXF_ERREUR(DIV_ZERO);
      }
      zone_x++; zone_y++;
    }
  }
  else
  {
    register double *zone_y = graphe->y;

    for (i = 0 ; i < graphe->nombre_point ; i++, zone_y++)
    {
      amp = SPL_tt_calc( correct, *zone_x, b, c, d);
      if (amp) /* Pas de risque de division par zro */
      {
	*zone_y /= amp;
      }
      else
      {
	EXF_ERREUR(DIV_ZERO);
      }
      zone_x++;
    }
  }

  free(b); free(c); free(d); /* On libre la mmoire */
  return VRAI;
}

/* ----------------- Correction de phase avant de faire la TF --------------------------------------------------\
 |   Renvoie FAUX en cas de problme                                                                            |
 |   Le graphe  corriger (dont on va faire la TF) DOIT tre de type complexe, mme si la fonction est relle   |
 |                                                                                                              |
 | graphe  : le graphe  corriger                                                                               |
 | correct : graphe contenant les phases servant  la correction                                                |
 | reel    : VRAI -> la partie imaginaire n'a pas t initialise -> suppose nulle (graphe rel  l'origine)   |
 \------------------------------------------------------------------------------------------------------------- */

boolean TF_corpha(Graphe *graphe, Graphe *correct, boolean reel)
{
  register Complexe *y;
  register double *x;
  double *b,*c,*d;
  double pha, reelle, imag;
  unsigned long i;

  if (graphe == NULL) return FAUX;
  if (correct == NULL) return FAUX;

  if ( (correct->x_min >= graphe->x_max ) || (correct->x_max <= graphe->x_min) )
  {
    form_alert(1,"[3][ Les intervalles en X | sont disjoints. | La correction est impossible ][ Annuler ]");
    return FAUX;
  }

  if (graphe->drapeaux.complexe == GR_NORMAL)
  {
    form_alert(1,"[3][ La correction de phase ne peut | se faire que sur un graphe complexe ][ Annuler ]");
    return FAUX;
  }

  if ( SPL_tt_prep( correct, &b, &c, &d) == -1 ) /* On prpare la Spline */
  {
    form_alert(1,"[3][ Problme de spline ][ Annuler ]");
    return FAUX;
  }

  x = graphe->x;
  y = (Complexe *) graphe->y;
  if (reel) /* Fonction relle -> partie imaginaire nulle */
  {
    for (i = 0 ; i < graphe->nombre_point ; i++)
    {
      pha = SPL_tt_calc( correct, *x, b, c, d);
      y->im  = - y->re * sin(pha);
      y->re *= cos(pha);
      x++; y++;
    }
  }
  else
  {
    double cosinus, sinus;

    for (i = 0 ; i < graphe->nombre_point ; i++)
    {
      pha = SPL_tt_calc( correct, *x, b, c, d);
      cosinus = cos(pha); sinus = sin(pha);
      reelle = y->re; imag = y->im;
      y->re = reelle * cosinus + imag   * sinus;
      y->im = imag   * cosinus - reelle * sinus;
      x++; y++;
    }
  }

  free(b); free(c); free(d); /* On libre la mmoire */
  return VRAI;
}

/* ~~~~~~~~ Correction des phases et amplitudes extraites par TFI ~~~~~~~~ *
 |                                                                         |
 | Correspond en fait au passage amplitude totale <-> fonction amplitude   |
 |                               phase totale     <-> dphasage            |
 |                                                                         |
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* ------------- Correction de la phase ---------------------------------------------------- */

Graphe *corr_phase(Graphe *graphe, double longueur)
{
  register unsigned long i;
  register double *zone_x, *zone_y;
  int retour;
  double decalage = 0;

  if (graphe == NULL) return NULL; /* N'a pas de sens sinon */

  if ( graphe->drapeaux.complexe != GR_NORMAL )
  {
    EXF_ERREUR(GRAPHE_COMPLEXE);
    return NULL;
  }

  if (longueur <= 0) /* Valeur irraliste */
  {
    retour = ext_formdo(CHOIX_PHASE);
    if (retour == PHA_ANNULE) return NULL;

    longueur = prend_fvaleur(CHOIX_PHASE,PHA_LONGUEUR);
    decalage = prend_fvaleur(CHOIX_PHASE,PHA_DECALAGE) * M_PI;
  }
	
  longueur *= 2;
  zone_x = graphe->x; zone_y = graphe->y;
  for (i = 0 ; i < graphe->nombre_point ; i++, zone_x++, zone_y++)
  {
    *zone_y -= longueur * *zone_x + decalage;
  }

  GRF_bornes(graphe, ORDONNE, DESORDONNE);
	
  return graphe;
}

/* ------------ Correction de l'amplitude -------------------------------------------------- *
 *      Renvoie le graphe si la correction a t effectue                                   */

Graphe *corr_amplitude(Graphe *graphe, double n_voisin, double longueur, double sigma)
{
  unsigned long i;
  register double *x, *y;
  int type_lpm;
  double gamma, eta;
  double *b = NULL, *c = NULL, *d = NULL;
  Graphe *libre_parcours = NULL;

  if (graphe == NULL) return NULL;

  if ( graphe->drapeaux.complexe != GR_NORMAL )
  {
    EXF_ERREUR(GRAPHE_COMPLEXE);
    return NULL;
  }

  if ( (n_voisin <= 0) || (sigma < 0) || (longueur <=0) ) /* Valeurs irralistes */
  {
    int retour;
    Graphe *g;

    for (i = 0, g = premier_graphe ; g != dernier_graphe ; g = g->suivant)
    {
      if ( (g->type == GR_LIBREP) || (g->type == GR_SSTYPE) )
	i++;
    }
    OBJ_active(CHOIX_AMPLITUDE, CORAMP_GRAPHE , i); /* Il existe des graphes de type libre parcours ou supposs tels */
    OBJ_active(CHOIX_AMPLITUDE, CORAMP_LPM_GRF, i);

    arbres[CHOIX_AMPLITUDE].fonction.arbre = GER_amplitude;

    do
    {
      retour = ext_formdo(CHOIX_AMPLITUDE);
      if (retour == AMP_ANNULE) return NULL; /* On annule le calcul */

      if (retour == CORAMP_GRAPHE)
      {
      }
    } while (retour != AMP_OK);

    n_voisin = prend_fvaleur(CHOIX_AMPLITUDE, VOISINS);
    sigma    = prend_fvaleur(CHOIX_AMPLITUDE, AMP_DEBYE);
    longueur = prend_fvaleur(CHOIX_AMPLITUDE, AMP_LONGUEUR);
  }

  if ( OBJ_choisit(CHOIX_AMPLITUDE, CORAMP_LPM_CT, QUESTION) )
  {
    type_lpm = CORAMP_LPM_CT;
    gamma = prend_fvaleur(CHOIX_AMPLITUDE, CORAMP_LPM_VAL);
  }
  else if ( OBJ_choisit(CHOIX_AMPLITUDE, CORAMP_LPM_VAR, QUESTION) )
  {
    type_lpm = CORAMP_LPM_VAR;
    gamma = prend_fvaleur(CHOIX_AMPLITUDE, CORAMP_LPM_VAL);
    eta  = prend_fvaleur(CHOIX_AMPLITUDE, CORAMP_VAL_ETA);
  }
  else
  {
    type_lpm = CORAMP_LPM_GRF;
    libre_parcours = (Graphe *) arbres[CHOIX_AMPLITUDE].pointeur;
    if (libre_parcours == NULL)
      type_lpm = PAS_DE_LPM;
    else
    {
      if (SPL_tt_prep( libre_parcours, &b, &c, &d) == -1) /* Erreur dans la Spline */
	type_lpm = PAS_DE_LPM;
    }
  }

  if (n_voisin == 0)
  {
    EXF_ERREUR(DIV_ZERO);
    return NULL;
  }

  /* Calcul de la correction */

  longueur *= longueur; /* On prend le carr */
  
  for (i = 0, x = graphe->x, y = graphe->y ;
       i < graphe->nombre_point ;
       i++, x++, y++)
  {
    double k = *x;
    if (k != 0) /* Le calcul a un sens */
    {
      double facteur = 1.0 * longueur * k / n_voisin; /* Amplitude du signal SANS multiplication par k^n... */
      if (sigma > 0)
      {
	facteur *= exp(2.0 * sigma * k * k); /* Facteur de Debye-Waller */
      }

      if (type_lpm == CORAMP_LPM_VAR) /* l(k) = 1 / [ G ( (n / k)^4 + k ) ] */
      {
	double inter = eta / k;
	inter *= inter;
	inter *= inter; /* Contient donc (eta / k)^4 */

	facteur *= exp( -2.0 * longueur * gamma * (k + inter) );
      }
      else if (type_lpm == CORAMP_LPM_CT) /* l(k) = 1 / G */
      {
	facteur *= exp( -2.0 * longueur * gamma );
      }
      else if (type_lpm == CORAMP_LPM_GRF)
      {
	double lambda = SPL_tt_calc(libre_parcours, k, b, c, d);

	if (lambda != 0)
	  facteur *= exp( -2.0 * longueur / lambda );
	else
	  EXF_ERREUR(DIV_ZERO);
      }

      *y *= facteur;
    }
  }

  GRF_bornes(graphe, ORDONNE, DESORDONNE);

  if (type_lpm == CORAMP_LPM_GRF)
  {
    free(b); free(c); free(d);
  }

  return graphe; /* Conversion a t effectue */
}

/* ------ Gestion de l'arbre associ  la correction d'amplitude extraite ------ */

static boolean GER_amplitude(int objet, int x, int y)
{
  switch (objet)
  {
  case AMP_OK:
  case AMP_ANNULE:
    return -objet;

  case CORAMP_LPM_CT: /* On demande correction de libre parcours moyen, constante */
    OBJ_choisit(CHOIX_AMPLITUDE, objet, FORCE_OUI);
    OBJ_active (CHOIX_AMPLITUDE, CORAMP_LPM_VAL, FORCE_OUI);
    OBJ_active (CHOIX_AMPLITUDE, CORAMP_VAL_ETA, FORCE_NON);
    OBJ_active (CHOIX_AMPLITUDE, CORAMP_GRAPHE , FORCE_NON);
    return VRAI;

  case CORAMP_LPM_VAR: /* On demande correction de libre parcours moyen, avec formule  deux variables */
    OBJ_choisit(CHOIX_AMPLITUDE, objet, FORCE_OUI);
    OBJ_active (CHOIX_AMPLITUDE, CORAMP_LPM_VAL, FORCE_OUI);
    OBJ_active (CHOIX_AMPLITUDE, CORAMP_VAL_ETA, FORCE_OUI);
    OBJ_active (CHOIX_AMPLITUDE, CORAMP_GRAPHE , FORCE_NON);
    return VRAI;

  case CORAMP_LPM_GRF: /* On demande correction de libre parcours moyen, dj calcul dans un graphe */
    OBJ_choisit(CHOIX_AMPLITUDE, objet, FORCE_OUI);
    OBJ_active (CHOIX_AMPLITUDE, CORAMP_LPM_VAL, FORCE_NON);
    OBJ_active (CHOIX_AMPLITUDE, CORAMP_VAL_ETA, FORCE_NON);
    OBJ_active (CHOIX_AMPLITUDE, CORAMP_GRAPHE , FORCE_OUI);
    return VRAI;

  case CORAMP_GRAPHE: /* Quel graphe de libre parcours utiliser pour la correction ? */
    arbres[CHOIX_AMPLITUDE].pointeur = (void *) GRF_choisir(GR_LIBREP);
    met_chaine(CHOIX_AMPLITUDE, CORAMP_GRAPHE, 30,
	       ( (Graphe *) (arbres[CHOIX_AMPLITUDE].pointeur) )->nom, TXT_COPIER);
    return VRAI;
  }

  return VRAI;
}
