/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998, Emmanuel CURIS - Cration de ce fichier: 23 juin 1998

   Module de filtrage des spectres (essentiellement par transforme de Fourier)

   Initialisation du calcul de la transforme de Fourier inverse
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "SYSTEME.H"

#include "EXAFS.H"
#include "ERREURS.H"
#include "GRAPHES.H"
#include "FOURIER.H"
#include "LISSAGE.H"
#include "SCIENCE.H"

/* ~~~~~~~~ Prparation dans le cas d'un calcul par algorithme rapide simple ~~~~~~~~~~
   Mthode : utilise l'intervalle demand pour le spectre d'arriv et le pas pour connatre le nombre de points
             calcule alors le R max correspondant
	     lisse et extrapole  zro le spectre dont on veut la TF ensuite           */

/* ---------------- Prparation pour la TF rapide ----------------------------------------------\
 |   Renvoie le graphe cr pour l'occasion                                                     |
 |           ainsi que le graphe * fentre * x^degre dans infos_tf->fenetre                     |
 |   Si graphe  traiter contient 2^k points : TFI directe                                      |
 |                       sinon               : on calcule les points qu'il faut                 |
 |                                                                                              |
 | infos_tf : informations ncessaires pour calculer la TF                                      |
 \--------------------------------------------------------------------------------------------- */

Graphe *TF_prep_rapide(TF_infos *infos, double r_min, double r_max, double pas)
{
  double r = 0;
  unsigned long i, nbr_pts; /* Nombre de points de la TF */
  int bit_1 = 0;
  Graphe *tf;
  register double *x;
  register Complexe *y;
  double x_fen = 0, pas_k;

  nbr_pts = infos->graphe->nombre_point; /* Nombre de points dans le spectre original */

  pas_k = infos->graphe->x_max - infos->graphe->x_min; /* Le pas en k (grossier) */
  pas_k /= (nbr_pts - 1);

  pas = prend_fvaleur(DEBUT_TF, FOUR_PAS); /* Valeur du pas dsir */
  if (pas <= 0)
  {
    EXF_ERREUR(PAS_FAUX);
    return NULL;
  }

  nbr_pts = (unsigned long) ( M_PI / (pas * pas_k) );
  bit_1 = 0;
  while ( (1 << bit_1) < nbr_pts)
    bit_1++;

  nbr_pts = 1 << bit_1; /* On arrondi  la valeur suprieure */

  pas_k = M_PI / ( pas * (double) nbr_pts ); /* Le pas en k que l'on va utiliser */

  /* Prparation */

  if ( infos->graphe->drapeaux.complexe == GR_NORMAL ) /* Graphe normal */
  {
    unsigned long i;
    double retour;
    double *a, *b, *c; /* Tableau pour le lissage par splines */
    double *nv_x;
    Complexe *nv_y;

    nv_x = (double *) calloc( nbr_pts, sizeof(double) );
    if (nv_x == NULL)
    {
      ERREUR(MANQUE_MEMOIRE);
      return NULL;
    }

    nv_y = (Complexe *) malloc( (size_t) (sizeof(Complexe) * nbr_pts) );
    if (nv_y == NULL)
    {
      free(nv_x);
      ERREUR(MANQUE_MEMOIRE);
      return NULL;
    }

    free(infos->fenetre->x); /* On libre les anciens points */
    free(infos->fenetre->y);
    infos->fenetre->x = nv_x;
    infos->fenetre->y = (double *) nv_y; /* On installe les nouveaux points */
    infos->fenetre->nombre_point = nbr_pts;

    /* On calcule la nouvelle courbe dont on veut la TF */
    retour = SPL_tt_prep(infos->graphe, &a, &b, &c);
    if (retour == -1) /* Erreur dans le calcul... */
      return NULL;

    x = nv_x;
    y = nv_y;
    retour = pas_k * sqrt( 2 / M_PI );
    for (i = 0 ; i < nbr_pts ; i++, x++, y++)
    {
      *x = x_fen; /* Valeur pour laquelle on calcule */
      if ( (x_fen < infos->graphe->x_min) || (x_fen > infos->graphe->x_max) )
	y->re = 0;
      else			/* On multiplie par la fentre au passage */
      {
	double y_fen = TF_calcfen(x_fen, infos->type_fenetre, infos->k_min, infos->k_max,
				  infos->k_1, infos->k_2, infos->param_fenetre, infos->symetrique);
	y->re = y_fen * SPL_tt_calc(infos->graphe, x_fen, a, b, c);

	y->re *= retour; /* Pour avoir une normalisation correcte */
      }
      y->im = 0;
	
      x_fen += pas_k;
    }

    if (infos->degre > 0) /* On veut multiplier par k^n */
    {
      int j;

      x = nv_x; y = nv_y;
      for (i = 0 ; i < nbr_pts ; i++, x++, y++)
      {
	x_fen = 1;

	for (j = 0 ; j < infos->degre ; j++)
	{
	  x_fen *= *x;
	}
	y->re *= x_fen;
      }
    }

    r_max = *(nv_x + nbr_pts - 1);
	
    free(a); free(b); free(c); /* Libration des tableaux de la spline, inutiles... */
  }
  else
  {
    /* Graphe complexe : pas encore fait... */
  }

  attend_evenement(MU_TIMER,          /* vnements attendus */
		   0, 0, 0, 0, 0,     /* rectangle 2 : aucun */
		   0,                 /* Compteur  1 -> pas de plantage sur TOS 1.0 */
		   &_rien_, &_rien_, &_rien_, &_rien_, (unsigned int *) (&_rien_), &_rien_);

  /* On prpare maintenant le graphe contenant la TF ... */

#ifdef _FRANCAIS_
  tf = GRF_creer(nbr_pts, GR_TF, GR_IM_MOD,
		 "TF (calcul en cours)", "Longueur de chemin", "Contribution", UNIT_ANG, UNIT_ARB);
#endif
#ifdef _ENGLISH_
  tf = GRF_creer(nbr_pts, GR_TF, GR_IM_MOD,
		 "Fourier transform (still computing)", "Path length", "Contribution", UNIT_ANG, UNIT_ARB);
#endif

  if (tf == NULL) return NULL;

  tf->drapeaux.protege = VRAI; /* Graphe en cours de calcul : on le protge... */
  tf->x_min = 0;
  tf->x_max = r_max;
  tf->y_min = -2; /*  l'usage, suffit le plus souvent */
  tf->y_max =  2;

  x = tf->x;
  y = (Complexe *) tf->y;
  for (i = 0 ; i < tf->nombre_point ; i++)
  {
    *x = r;
    y->re = 0; y->im = 0; /* On met  zro la TF - juste pour le dessin... */

    r += pas; y++; x++;
  }

  if ( ( prend_fvaleur(DEBUT_TF, R_MAX) <= 0 ) || ( prend_fvaleur(DEBUT_TF, R_MAX) > r_max ) )
  {
    /* On s'assure que l'on ne demande pas de faire mieux que le max. thorique... */

    printf("Valeur de R_max incorrecte : %f, max. %f. Correction automatique...\n",
	   prend_fvaleur(DEBUT_TF, R_MAX), r_max);
    met_fvaleur(DEBUT_TF, R_MAX, r_max);
  }

  return tf;
}

/* ~~~~~~~~ Prparation dans le cas d'un calcul par la mthode brutale ~~~~~~~~~~~~~~~~ */

/* -------------------- Prparation du graphe qui contiendra la TF -----------------------------\
 |  Renvoie le graphe dsir (NULL si erreur)                                                   |
 |          ainsi que le graphe de dpart * fentre * x^degre dans infos_tf->fenetre            |
 |                                                                                              |
 |  infos_tf : informations ncessaires  la TF                                                 |
 \--------------------------------------------------------------------------------------------- */

Graphe *TF_prep_normal(TF_infos *infos_tf, double r_min, double r_max, double pas)
{
  double r; /* Intervalle et pas pour la fonction d'arrive */
  unsigned long nbr_pts, i;
  register double *x;
  register Complexe *y;
  Graphe *fourier;

  if (r_max == r_min)
  {
    ERREUR(VALEURS_EGALES);
    return NULL;			
  }
  if (r_max < r_min) /* On change les deux */
  {
    double r = r_max; r_max = r_min; r_min = r;
  }

  if ( (pas > (r_max - r_min)) || (pas <= 0) )
  {
    EXF_ERREUR(PAS_FAUX);
    return NULL;
  }

  nbr_pts = 1 + (unsigned long) ( (r_max - r_min) / pas ); /* Nombre de points de la TF */

#ifdef _FRANCAIS_
  fourier = GRF_creer(nbr_pts, GR_TF, GR_IM_MOD, "TF (calcul en cours)",
		      "Longueur de chemin", "Contribution", UNIT_ANG, UNIT_ARB);
#endif
#ifdef _ENGLISH_
  fourier = GRF_creer(nbr_pts, GR_TF, GR_IM_MOD, "Fourier Transform (still computing)",
		      "Path length", "Contribution", UNIT_ANG, UNIT_ARB);
#endif

  if (fourier == NULL) return NULL;

  fourier->drapeaux.protege = VRAI; /* Graphe en cours de calcul : on le protge... */
  fourier->x_min = r_min;
  fourier->x_max = r_max;
  fourier->y_min = -2; /*  l'usage, suffit le plus souvent */
  fourier->y_max =  2;

  r = r_min;
  x = fourier->x;
  y = (Complexe *) fourier->y;
  for (i = 0 ; i < nbr_pts ; i++)
  {
    *x = r;
    y->re = 0; y->im = 0; /* On met  zro la TF */
    
    r += pas; y++; x++;
  }

  /* Prparation, dans tf_fenetre, du graphe dont on veut la TF */

  if ( infos_tf->graphe->drapeaux.complexe == GR_NORMAL )
  {
    register Complexe *y_fen;
    register double *x, *y;
    int j;

    y_fen = (Complexe *) infos_tf->fenetre->y;
    x = infos_tf->graphe->x;
    y = infos_tf->graphe->y;
    for (i = 0 ; i < infos_tf->graphe->nombre_point ; i++, x++, y++, y_fen++)
    {
      double x_tf = *x, y_tf = *y;
      for (j = 1 ; j <= infos_tf->degre ; j++)
      {
	y_tf *= x_tf;
      }
      y_fen->re *= y_tf; /* Spectre rel -> seule la partie relle est modifie */
      y_fen->im  = 0;    /* Pour tre tranquille */
    }
  }
  else /* Spectre de dpart : spectre complexe */
  {
    register Complexe *y_fen, *y;
    Complexe y_tf;
    register double *x;
    unsigned int j;

    y_fen = (Complexe *) infos_tf->fenetre->y;
    x = infos_tf->graphe->x;
    y = (Complexe *) (infos_tf->graphe->y);
    for (i = 0 ; i < infos_tf->graphe->nombre_point ; i++, x++, y++, y_fen++)
    {
      double x_tf = *x;
      
      y_tf = *y;
      for (j = 1 ; j <= infos_tf->degre ; j++)
      {
	y_tf.re *= x_tf;
	y_tf.im *= x_tf;
      }

      y_fen->im  = y_fen->re * y_tf.im;
      y_fen->re *= y_tf.re;
    }
  }

  return fourier; /* Le graphe qui contiendra la TF */
}
