/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998, Emmanuel CURIS - Cration de ce fichier: 23 juin 1998

   Calcul des coefficients d'une spline approchant le graphe indiqu
   Cas d'un graphe complexe
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <SYSTEME.H>

#include <EXAFS.H>
#include <GRAPHES.H>
#include <ERREURS.H>
#include <LISSAGE.H>

/*********** Lissage par une "spline", la plus adapte possible ********************************
 *  Suppose que le graphe  lisser soit tri, par ordre croissant des X
 *          que la liste de zones soit trie (mme ordre) et disjointe
 ***********************************************************************************************/

/* --------- Prpare les tableaux de coefficients utiliss pour les splines ---------------------\
 |  Renvoie pas grand chose, -1 si erreur                                                        |
 |  Version pour une srie de portions du graphe seulement                                       |
 |  Les quatre tableaux crs dans la procdure ne sont PAS librs : ne pas oublier de le faire |
 |                                                                                               |
 | graphe     : graphe  lisser                                                                  |
 | nbr_zones  : nombre d'intervalles de points valables                                          |
 | zones      : bornes de ces intervalles ([0] = min1, [1] = max1, [2] = min2, ...)              |
 | b, c, d    : tableaux des coefficients (rservs DANS la procdure, d'o leur type...)        |
 | tab_x      : tableau des X de la portion de graphe rellement utilise, cr sur place        |
 | tab_y      : tableau des Y de la portion de graphe rellement utilise, cr sur place        |
 | nbr_points : nombre de points dans le tableau ci-dessus                                       |
 \---------------------------------------------------------------------------------------------- */
 
double CSPL_prepare(Graphe *graphe, int nbr_zones, double *zones, 
	 	    Complexe **b, Complexe **c, Complexe **d, double **tab_x, Complexe **tab_y,
		    unsigned long *nbr_points)
{
  int a;
  double x_min, x_max; /* Intervalle total de lissage */
  register double *zone_x;
  register Complexe *zone_y;
  unsigned long i, nbr_pts = 0;

  if ( (nbr_zones == 0) || (zones == NULL) ) return -1; /* Aucun intervalle spcifi */

  x_min = zones[0]; x_max = zones[1];
  for (a = 2 ; a <= 2*nbr_zones ; a++)
  {
    if (zones[a] < x_min) x_min = zones[a];
    if (zones[a] > x_max) x_max = zones[a];
  }

  /* Calcul du nombre de points */
  for (i = 0, zone_x = graphe->x ; i < graphe->nombre_point ; i++, zone_x++)
  {
    int a;
    double x = *zone_x;
    for (a = 0 ; a < 2*nbr_zones ; a += 2)
    {
      if ( (x >= zones[a]) && (x <= zones[a + 1]) )
      {
	nbr_pts++;
	break;
      }
    }
  }

  if (nbr_pts < 4) /* Moins de quatre -> sans intrt */
  {
    LIS_ERREUR(SPLINE_INADAPTE);
    return -SPLINE_INADAPTE;
  }

  *tab_x = (double *)   calloc( nbr_pts, sizeof(double) );
  *tab_y = (Complexe *) calloc( nbr_pts, sizeof(Complexe) );
  if ( (*tab_x == NULL) || (*tab_y == NULL) )
  {
    free(*tab_x); free(*tab_y);
    ERREUR(MANQUE_MEMOIRE);
    return -MANQUE_MEMOIRE;
  }

  *b = (Complexe *) malloc( sizeof(Complexe) * nbr_pts );
  *c = (Complexe *) malloc( sizeof(Complexe) * nbr_pts );
  *d = (Complexe *) malloc( sizeof(Complexe) * nbr_pts );
  if ( (d == NULL) || (c == NULL) || (b == NULL) )
  {
    free(*tab_x); free(*tab_y);
    free(*b); free(*c); free(*d); /* On libre la mmoire ventuellement prise */
    ERREUR(MANQUE_MEMOIRE);
    return -MANQUE_MEMOIRE;                 /* Mmoire insuffisante                      */
  }

  /* On recopie la partie intressante du graphe dans tableau */
  
  zone_x = *tab_x;
  zone_y = *tab_y;
  for (i = 0 ; i < graphe->nombre_point ; i++)
  {
    int a;
    double x = *(graphe->x + i);
    for (a = 0 ; a < 2*nbr_zones ; a += 2)
    {
      if ( (x >= zones[a]) && (x <= zones[a + 1]) )
      {
	*zone_x = x;
	*zone_y = *( ( (Complexe *) graphe->y ) + i);
	zone_x++;
	zone_y++;
	break; /* Ne sort que d'un seul niveau -> on passe  la boucle for (i...) */
      }
    }
  }

  /* Prparation des tableaux de coefficients */
  zone_x = *tab_x;
  zone_y = *tab_y;

  if ( zone_x[1] == zone_x[0] )
  {
    EXF_ERREUR(ERR_DOUBLONS);
    free(*b); free(*c); free(*d);
    free(*tab_x); free(*tab_y);
    return -ERR_DOUBLONS;
  }
  (*d)[0].re = zone_x[1] - zone_x[0];
  (*d)[0].im = 0;

  (*c)[1].re = ( zone_y[1].re - zone_y[0].re ) / (*d)[0].re; /* d est rel */
  (*c)[1].im = ( zone_y[1].im - zone_y[0].im ) / (*d)[0].re;

  for (i = 1 ; i < nbr_pts - 1 ; i++)
  {
    if ( zone_x[i + 1] == zone_x[i] )
    {
      EXF_ERREUR(ERR_DOUBLONS);
      free(*b); free(*c); free(*d);
      free(*tab_x); free(*tab_y);
      return -ERR_DOUBLONS;
    }
    (*d)[i].re = zone_x[i + 1] - zone_x[i];
    (*d)[i].im = 0; /* d[i] = x[i+1] - x[i], donc est rel : les X sont rels */

    (*b)[i].re = 2 * ( (*d)[i - 1].re + (*d)[i].re );
    (*b)[i].im = 0; /* b[i] = 2 * ( d[i-1] + d[i] ) est rel : les d[i] sont tous rels */

    (*c)[i + 1].re = ( zone_y[i + 1].re - zone_y[i].re ) / (*d)[i].re;
    (*c)[i + 1].im = ( zone_y[i + 1].im - zone_y[i].im ) / (*d)[i].re;
    /* c[i] = (y[i+1] - y[i]) / d[i] : complexe - d[i] est rel */

    (*c)[i].re = (*c)[i + 1].re - (*c)[i].re;
    (*c)[i].im = (*c)[i + 1].im - (*c)[i].im;
    /* c[i] = c[i + 1] - c[i] */
  }
  (*b)[0].re = -(*d)[0].re;
  (*b)[nbr_pts - 1].re = -(*d)[nbr_pts - 2].re;
		
  if ( (zone_x[1] == zone_x[3]) || (zone_x[2] == zone_x[0]) || (zone_x[3] == zone_x[0]) )
  {
    EXF_ERREUR(ERR_DOUBLONS);
    free(*b); free(*c); free(*d);
    free(*tab_x); free(*tab_y);
    return -1;
  }
  (*c)[0].re  = (*c)[2].re / (zone_x[3] - zone_x[1])
              - (*c)[1].re / (zone_x[2] - zone_x[0]);
  (*c)[0].im  = (*c)[2].im / (zone_x[3] - zone_x[1])
              - (*c)[1].im / (zone_x[2] - zone_x[0]);

  (*c)[0].re *= (*d)[0].re * (*d)[0].re / (zone_x[3] - zone_x[0]);
  (*c)[0].im *= (*d)[0].re * (*d)[0].re / (zone_x[3] - zone_x[0]);

  if ( (zone_x[nbr_pts - 1] == zone_x[nbr_pts - 3]) 
    || (zone_x[nbr_pts - 2] == zone_x[nbr_pts - 4])
    || (zone_x[nbr_pts - 4] == zone_x[nbr_pts - 1]) )
  {
    EXF_ERREUR(ERR_DOUBLONS);
    free(*b); free(*c); free(*d);
    free(*tab_x); free(*tab_y);
    return -ERR_DOUBLONS;
  }
  (*c)[nbr_pts - 1].re = (*c)[nbr_pts - 2].re / (zone_x[nbr_pts - 1] - zone_x[nbr_pts - 3])
                       - (*c)[nbr_pts - 3].re / (zone_x[nbr_pts - 2] - zone_x[nbr_pts - 4]);
  (*c)[nbr_pts - 1].im = (*c)[nbr_pts - 2].im / (zone_x[nbr_pts - 1] - zone_x[nbr_pts - 3])
                       - (*c)[nbr_pts - 3].im / (zone_x[nbr_pts - 2] - zone_x[nbr_pts - 4]);
  (*c)[nbr_pts - 1].re *= - (*d)[nbr_pts - 2].re * (*d)[nbr_pts - 2].re
                        / (zone_x[nbr_pts - 1] - zone_x[nbr_pts - 4]);
  (*c)[nbr_pts - 1].im *= - (*d)[nbr_pts - 2].re * (*d)[nbr_pts - 2].re
                        / (zone_x[nbr_pts - 1] - zone_x[nbr_pts - 4]);

  for (i = 1 ; i < nbr_pts ; i ++)
  {
    double pas;

    if ( (*b)[i - 1].re == 0 )	/* b est rel,  cette tape -> Re(b) = 0 <=> b = 0 */
    {
      EXF_ERREUR(DIV_ZERO);
      free(*b); free(*c); free(*d);
      free(*tab_x); free(*tab_y);
      return -DIV_ZERO;
    }

    pas = (*d)[i - 1].re / (*b)[i - 1].re; /* pas = d[i-1]/b[i-1] est rel : d et b sont rels */
    (*b)[i].re -= pas * (*d)[i - 1].re; /* b[i] = b[i] - pas * d[i-1] est rel */

    (*c)[i].re -= pas * (*c)[i - 1].re; /* c, en revanche, est vraiment complexe */
    (*c)[i].im -= pas * (*c)[i - 1].im; /* c[i] = c[i] - pas * c[i-1] */
  }
  if ( (*b)[nbr_pts - 1].re == 0 ) /* b est rel -> b est nul <=> Re(b) = 0 */
  {
    EXF_ERREUR(DIV_ZERO);
    free(*b); free(*c); free(*d);
    free(*tab_x); free(*tab_y);
    return -DIV_ZERO;
  }
  (*c)[nbr_pts - 1].re /= (*b)[nbr_pts - 1].re; /* b est toujours rel */
  (*c)[nbr_pts - 1].im /= (*b)[nbr_pts - 1].re;

  /* b et d dj verifis au-dessus -> pas la peine de recommencer */
  for (i = nbr_pts - 2 ; i > 0 ; i--)
  {
    (*c)[i].re -= (*d)[i].re * (*c)[i + 1].re; /* c[i] = c[i] - d[i] * c[i+1] */
    (*c)[i].im -= (*d)[i].re * (*c)[i + 1].im; /* d est toujours rel */
    (*c)[i].re /= (*b)[i].re; /* c[i] = c[i]/b[i], b est rel */
    (*c)[i].im /= (*b)[i].re;
  }
  (*c)[0].re -= (*d)[0].re * (*c)[1].re;
  (*c)[0].im -= (*d)[0].re * (*c)[1].im;
  (*c)[0].re /= (*b)[0].re;
  (*c)[0].im /= (*b)[0].re;

  (*b)[nbr_pts - 1].re = (zone_y[nbr_pts - 1].re - zone_y[nbr_pts - 2].re ) / (*d)[nbr_pts - 2].re;
  (*b)[nbr_pts - 1].im = (zone_y[nbr_pts - 1].im - zone_y[nbr_pts - 2].im ) / (*d)[nbr_pts - 2].re;

  (*b)[nbr_pts - 1].re += (*d)[nbr_pts - 2].re * ((*c)[nbr_pts - 2].re + 2 * (*c)[nbr_pts - 1].re );
  (*b)[nbr_pts - 1].im += (*d)[nbr_pts - 2].re * ((*c)[nbr_pts - 2].im + 2 * (*c)[nbr_pts - 1].im );

  for (i = 0 ; i < nbr_pts - 1 ; i++)
  {
    (*b)[i].re = (zone_y[i + 1].re - zone_y[i].re) / (*d)[i].re; /* b[i] = (y[i+1] - y[i])/d[i] */
    (*b)[i].im = (zone_y[i + 1].im - zone_y[i].im) / (*d)[i].re; /* d est toujours rel */

    (*b)[i].re -= (*d)[i].re * ( (*c)[i + 1].re + 2 * (*c)[i].re );
    (*b)[i].im -= (*d)[i].re * ( (*c)[i + 1].im + 2 * (*c)[i].im );
    /* b[i] = b[i] - d[i] * ( c[i+1] + 2*c[i]] ), d[i] est encore rel  ce moment l */

    (*d)[i].im  = ( (*c)[i + 1].im - (*c)[i].im ) / (*d)[i].re; /* d[i] = (c[i+1] - c[i])/d[i] */
    (*d)[i].re  = ( (*c)[i + 1].re - (*c)[i].re ) / (*d)[i].re; /* Im d'abord car d rel... */

    (*c)[i].re *= 3; /* c[i] = 3 * c[i] */
    (*c)[i].im *= 3;
  }
  (*c)[nbr_pts - 1].re *= 3;
  (*c)[nbr_pts - 1].im *= 3;
  (*d)[nbr_pts - 1] = (*d)[nbr_pts - 2];
    
  *nbr_points = nbr_pts;

  return 1; /* Pas de problme */
}

/* --------- Prpare les tableaux de coefficients utiliss pour les splines ---------------------\
 |  Renvoie pas grand chose, -code_erreur si erreur, 0 si ok
 |  Version pour _tout_ le graphe pass en paramtre                                             |
 |  Les quatre tableaux crs dans la procdure ne sont PAS librs : ne pas oublier de le faire |
 |
 | graphe     : graphe  lisser                                                                  |
 | b, c, d    : tableaux des coefficients (rservs DANS la procdure, d'o leur type...)        |
 \---------------------------------------------------------------------------------------------- */
 
double CSPL_tt_prep(Graphe *graphe, Complexe **b, Complexe **c, Complexe **d)
{
  register double   *zone_x;
  register Complexe *zone_y;
  unsigned long i, nbr_pts;

  if (graphe->nombre_point < 4) /* Moins de quatre -> sans intrt */
  {
    LIS_ERREUR(SPLINE_INADAPTE);
    return -SPLINE_INADAPTE;
  }

  nbr_pts = graphe->nombre_point; /* Pour viter de se prendre la tte aprs... */

  *b = (Complexe *) malloc( sizeof(Complexe) * nbr_pts );
  *c = (Complexe *) malloc( sizeof(Complexe) * nbr_pts );
  *d = (Complexe *) malloc( sizeof(Complexe) * nbr_pts );
  if ( (d == NULL) || (c == NULL) || (b == NULL) )
  {
    free(*b); free(*c); free(*d); /* On libre la mmoire ventuellement prise */
    ERREUR(MANQUE_MEMOIRE);
    return -MANQUE_MEMOIRE;                 /* Mmoire insuffisante                      */
  }

  /* Prparation des tableaux de coefficients */

  zone_x = graphe->x;
  zone_y = (Complexe *) graphe->y;

  if ( zone_x[1] == zone_x[0] )
  {
    EXF_ERREUR(ERR_DOUBLONS);
    free(*b); free(*c); free(*d);
    return -ERR_DOUBLONS;
  }
  (*d)[0].re = zone_x[1] - zone_x[0];
  (*d)[0].im = 0;

  (*c)[1].re = ( zone_y[1].re - zone_y[0].re ) / (*d)[0].re; /* d est rel */
  (*c)[1].im = ( zone_y[1].im - zone_y[0].im ) / (*d)[0].re;

  for (i = 1 ; i < nbr_pts - 1 ; i++)
  {
    if ( zone_x[i + 1] == zone_x[i] )
    {
      EXF_ERREUR(ERR_DOUBLONS);
      free(*b); free(*c); free(*d);
      return -ERR_DOUBLONS;
    }
    (*d)[i].re = zone_x[i + 1] - zone_x[i];
    (*d)[i].im = 0; /* d[i] = x[i+1] - x[i], donc est rel : les X sont rels */

    (*b)[i].re = 2 * ( (*d)[i - 1].re + (*d)[i].re );
    (*b)[i].im = 0; /* b[i] = 2 * ( d[i-1] + d[i] ) est rel : les d[i] sont tous rels */

    (*c)[i + 1].re = ( zone_y[i + 1].re - zone_y[i].re ) / (*d)[i].re;
    (*c)[i + 1].im = ( zone_y[i + 1].im - zone_y[i].im ) / (*d)[i].re;
    /* c[i] = (y[i+1] - y[i]) / d[i] : complexe - d[i] est rel */

    (*c)[i].re = (*c)[i + 1].re - (*c)[i].re;
    (*c)[i].im = (*c)[i + 1].im - (*c)[i].im;
    /* c[i] = c[i + 1] - c[i] */
  }
  (*b)[0].re = -(*d)[0].re;
  (*b)[nbr_pts - 1].re = -(*d)[nbr_pts - 2].re;
		
  if ( (zone_x[1] == zone_x[3]) || (zone_x[2] == zone_x[0]) || (zone_x[3] == zone_x[0]) )
  {
    EXF_ERREUR(ERR_DOUBLONS);
    free(*b); free(*c); free(*d);
    return -ERR_DOUBLONS;
  }
  (*c)[0].re  = (*c)[2].re / (zone_x[3] - zone_x[1])
              - (*c)[1].re / (zone_x[2] - zone_x[0]);
  (*c)[0].im  = (*c)[2].im / (zone_x[3] - zone_x[1])
              - (*c)[1].im / (zone_x[2] - zone_x[0]);

  (*c)[0].re *= (*d)[0].re * (*d)[0].re / (zone_x[3] - zone_x[0]);
  (*c)[0].im *= (*d)[0].re * (*d)[0].re / (zone_x[3] - zone_x[0]);

  if ( (zone_x[nbr_pts - 1] == zone_x[nbr_pts - 3]) 
    || (zone_x[nbr_pts - 2] == zone_x[nbr_pts - 4])
    || (zone_x[nbr_pts - 4] == zone_x[nbr_pts - 1]) )
  {
    EXF_ERREUR(ERR_DOUBLONS);
    free(*b); free(*c); free(*d);
    return -ERR_DOUBLONS;
  }
  (*c)[nbr_pts - 1].re = (*c)[nbr_pts - 2].re / (zone_x[nbr_pts - 1] - zone_x[nbr_pts - 3])
                       - (*c)[nbr_pts - 3].re / (zone_x[nbr_pts - 2] - zone_x[nbr_pts - 4]);
  (*c)[nbr_pts - 1].im = (*c)[nbr_pts - 2].im / (zone_x[nbr_pts - 1] - zone_x[nbr_pts - 3])
                       - (*c)[nbr_pts - 3].im / (zone_x[nbr_pts - 2] - zone_x[nbr_pts - 4]);

  (*c)[nbr_pts - 1].re *= - (*d)[nbr_pts - 2].re * (*d)[nbr_pts - 2].re
                        / (zone_x[nbr_pts - 1] - zone_x[nbr_pts - 4]);
  (*c)[nbr_pts - 1].im *= - (*d)[nbr_pts - 2].re * (*d)[nbr_pts - 2].re
                        / (zone_x[nbr_pts - 1] - zone_x[nbr_pts - 4]);

  for (i = 1 ; i < nbr_pts ; i ++)
  {
    double pas;

    if ( (*b)[i - 1].re == 0 )	/* b est rel,  cette tape -> Re(b) = 0 <=> b = 0 */
    {
      EXF_ERREUR(DIV_ZERO);
      free(*b); free(*c); free(*d);
      return -DIV_ZERO;
    }

    pas = (*d)[i - 1].re / (*b)[i - 1].re; /* pas = d[i-1]/b[i-1] est rel : d et b sont rels */
    (*b)[i].re -= pas * (*d)[i - 1].re; /* b[i] = b[i] - pas * d[i-1] est rel */

    (*c)[i].re -= pas * (*c)[i - 1].re; /* c, en revanche, est vraiment complexe */
    (*c)[i].im -= pas * (*c)[i - 1].im; /* c[i] = c[i] - pas * c[i-1] */
  }
  if ( (*b)[nbr_pts - 1].re == 0 ) /* b est rel -> b est nul <=> Re(b) = 0 */
  {
    EXF_ERREUR(DIV_ZERO);
    free(*b); free(*c); free(*d);
    return -DIV_ZERO;
  }
  (*c)[nbr_pts - 1].re /= (*b)[nbr_pts - 1].re; /* b est toujours rel */
  (*c)[nbr_pts - 1].im /= (*b)[nbr_pts - 1].re;

  /* b et d dj verifis au-dessus -> pas la peine de recommencer */
  for (i = nbr_pts - 2 ; i > 0 ; i--)
  {
    (*c)[i].re -= (*d)[i].re * (*c)[i + 1].re; /* c[i] = c[i] - d[i] * c[i+1] */
    (*c)[i].im -= (*d)[i].re * (*c)[i + 1].im; /* d est toujours rel */
    (*c)[i].re /= (*b)[i].re; /* c[i] = c[i]/b[i], b est rel */
    (*c)[i].im /= (*b)[i].re;
  }
  (*c)[0].re -= (*d)[0].re * (*c)[1].re;
  (*c)[0].im -= (*d)[0].re * (*c)[1].im;
  (*c)[0].re /= (*b)[0].re;
  (*c)[0].im /= (*b)[0].re;

  (*b)[nbr_pts - 1].re = (zone_y[nbr_pts - 1].re - zone_y[nbr_pts - 2].re ) / (*d)[nbr_pts - 2].re;
  (*b)[nbr_pts - 1].im = (zone_y[nbr_pts - 1].im - zone_y[nbr_pts - 2].im ) / (*d)[nbr_pts - 2].re;

  (*b)[nbr_pts - 1].re += (*d)[nbr_pts - 2].re * ((*c)[nbr_pts - 2].re + 2 * (*c)[nbr_pts - 1].re );
  (*b)[nbr_pts - 1].im += (*d)[nbr_pts - 2].re * ((*c)[nbr_pts - 2].im + 2 * (*c)[nbr_pts - 1].im );

  for (i = 0 ; i < nbr_pts - 1 ; i++)
  {
    (*b)[i].re = (zone_y[i + 1].re - zone_y[i].re) / (*d)[i].re; /* b[i] = (y[i+1] - y[i])/d[i] */
    (*b)[i].im = (zone_y[i + 1].im - zone_y[i].im) / (*d)[i].re; /* d est toujours rel */

    (*b)[i].re -= (*d)[i].re * ( (*c)[i + 1].re + 2 * (*c)[i].re );
    (*b)[i].im -= (*d)[i].re * ( (*c)[i + 1].im + 2 * (*c)[i].im );
    /* b[i] = b[i] - d[i] * ( c[i+1] + 2*c[i]] ), d[i] est encore rel  ce moment l */

    (*d)[i].im  = ( (*c)[i + 1].im - (*c)[i].im ) / (*d)[i].re; /* d[i] = (c[i+1] - c[i])/d[i] */
    (*d)[i].re  = ( (*c)[i + 1].re - (*c)[i].re ) / (*d)[i].re; /* Im d'abord car d rel... */

    (*c)[i].re *= 3; /* c[i] = 3 * c[i] */
    (*c)[i].im *= 3;
  }
  (*c)[nbr_pts - 1].re *= 3;
  (*c)[nbr_pts - 1].im *= 3;
  (*d)[nbr_pts - 1] = (*d)[nbr_pts - 2];

  return 0; /* Pas de problme */
}

/* -------- Conversion arrte ici faute de papier -------- */

/* --------- Calcule un point de la spline obtenue par la procdure prcdente -----------\
 |   Renvoie la valeur en ce point (0 en cas d'erreur...)
 |   Version pour une spline calcule sur plusieurs graphes
 |   Suppose de plus que tableau est class par X croissant
 |
 | x       : point en lequel calculer
 | nbr_pts : nombre de points utiliss pour le calcul de la spline
 | tableau : tableaux des X et Y (tous les X puis tous les Y), suppos tri par X croissant
 | b, c, d : coefficients de la spline calculs par la fonction au-dessus
 \--------------------------------------------------------------------------------------- */

Complexe CSPL_calcule(double x, unsigned long nbr_pts, double *tab_x, Complexe *tab_y,
		      Complexe *b, Complexe *c, Complexe *d)
{
  Complexe renvoi = {0, 0};
  double delta_x;
  unsigned long i;

  if (tab_x == NULL) return renvoi;
  if (tab_y == NULL) return renvoi;
  if (b == NULL) return renvoi;
  if (c == NULL) return renvoi;
  if (d == NULL) return renvoi;

  /* On recherche le point en x le plus proche, par dichotomie */
  if ( x <= tab_x[0])
  {
    delta_x = x - tab_x[0];
    i = 0;
  }
  else if (x >= tab_x[nbr_pts - 1])
  {
    delta_x = x - tab_x[nbr_pts - 1];
    i = nbr_pts - 1;
  }
  else /* if ( (x > tableau[1]) || ( x < tableau[0]) ) */
  {
    /* Recherche du point le plus proche */

    i = 1;
    while (x > tab_x[i])
      i++; /* lment qui suit juste x */
    i--; /* On prend l'lment juste avant */
    
    delta_x = x - tab_x[i];
  }

  if (fabs(delta_x) > 0)
  {
    renvoi.re  = delta_x * d[i].re;
    renvoi.re += c[i].re;
    renvoi.re *= delta_x;
    renvoi.re += b[i].re;
    renvoi.re *= delta_x;

    renvoi.im  = delta_x * d[i].im;
    renvoi.im += c[i].im;
    renvoi.im *= delta_x;
    renvoi.im += b[i].im;
    renvoi.im *= delta_x;
  }
  renvoi.re += tab_y[i].re; /* R = Y + b * X + c * X^2 + d * X^3 */
  renvoi.im += tab_y[i].im; /* les X sont rels, donc c'est ok */
  
  return renvoi;
}

/* --------- Calcule un point de la spline obtenue par la procdure prcdente -----------\
 |   Renvoie la valeur en ce point (0 en cas d'erreur, mais avec une alerte...)
 |   Version pour une spline sur tout un graphe
 |
 | graphe  : graphe qui a servi pour le calcul de la spline
 | x       : point en lequel calculer
 | b, c, d : coefficients de la spline calculs par la fonction au-dessus
 \--------------------------------------------------------------------------------------- */

Complexe CSPL_tt_calc(Graphe *graphe, double x, Complexe *b, Complexe *c, Complexe *d)
{
  Complexe renvoi = {0, 0};
  double delta_x;
  unsigned long i;

  if (b == NULL) return renvoi; /* Problme -> donner une alerte */
  if (c == NULL) return renvoi;
  if (d == NULL) return renvoi;
  
  /* On recherche le point en x le plus proche, par dichotomie */
  if ( x <= graphe->x_min)
  {
    delta_x = x - graphe->x_min;
    i = 0;
  }
  else if (x >= graphe->x_max)
  {
    delta_x = x - graphe->x_max;
    i = graphe->nombre_point - 1;
  }
  else
  {
    /* Recherche du point le plus proche */

    i = 1;
    while (x > (graphe->x)[i])
      i++; /* lment qui suit juste x */
    i--; /* On prend l'lment juste avant */
    
    delta_x = x - (graphe->x)[i];
  }

  if (fabs(delta_x) > 0)
  {
    renvoi.re  = delta_x * d[i].re;
    renvoi.re += c[i].re;
    renvoi.re *= delta_x;
    renvoi.re += b[i].re;
    renvoi.re *= delta_x;

    renvoi.im  = delta_x * d[i].im;
    renvoi.im += c[i].im;
    renvoi.im *= delta_x;
    renvoi.im += b[i].im;
    renvoi.im *= delta_x;
  }
  renvoi.re += ( (Complexe *) (graphe->y) )[i].re;
  renvoi.im += ( (Complexe *) (graphe->y) )[i].im;
  /* R = Y + b * X + c * X^2 + d * X^3, avec les X rels */

  return renvoi;
}
