/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-1998, Emmanuel CURIS
   Ce fichier : cration le 28 septembre 1998

   Module d'ajustement par un modle  n couches

   Calcul d'un signal EXAFS (et associ)  partir des rsultats d'ajustement
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "SYSTEME.H"

#include "EXAFS.H"
#include "GRAPHES.H"
#include "LISSAGE.H"
#include "SCIENCE.H"

#include "couches.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int exafs_1_couche(Couche *couche, Graphe *exafs);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Calcul du signal exafs associ  l'ensemble des couches -------\
 |  Renvoie : 0 si tout s'est bien pass, un code d'erreur sinon           |
 |                                                                         |
 | prem_couche : premire couche du modle                                 |
 | exafs       : le spectre contenant le rsultat                          |
 |               (le signal est calcul pour chacun des x=k de ce graphe)  |
 \------------------------------------------------------------------------ */

int calc_exafs_ncouches(Couche *prem_couche, Graphe *exafs)
{
  unsigned long i;
  register double *x, *y;
  Couche *couche;
  double s0_2;

  if ( (exafs == NULL) || (prem_couche == NULL) )
    return -1; /* Pas d'entre ! */

  if ( exafs->drapeaux.complexe != GR_NORMAL ) /* Le spectre doit tre de type normal */
    return -2;

  /* On initialise  zro les y du graphe */

  for (i = 0, y = exafs->y ; i < exafs->nombre_point ; i++, y++)
    *y = 0;

  /* On calcule pour chacune des couches utilise du modle la contribution */

  for (couche = prem_couche ; couche ; couche = couche->suivante)
  {
    if ( couche->utilisee == VRAI ) /* On utilise la couche pour le modle */
    {
      int retour;

      retour = exafs_1_couche(couche, exafs);
      if (retour) /* Il y a eu un problme */
	return retour;
    }
  }

  /* On calcule maintenant pour les paramtres gnraux : multiplication par S0^2 et pondration par k^n */

  s0_2 = VAR_prend_valeur( &(S0_2) );

  for (i = 0, x = exafs->x, y = exafs->y ; i < exafs->nombre_point ; i++, x++, y++)
  {
    *y *= s0_2;

    if ( ponder_ajuste > 0 ) /* Pondration par k^n */
    {
      double k_p_n = *x;
      int i;

      for (i = 0 ; i < ponder_ajuste ; i++)
	k_p_n *= *x;

      *y *= k_p_n;
    }
  }

  GRF_bornes(exafs, ORDONNE, DESORDONNE); /* Pour tre sr de l'affichage... */

  return 0; /* Pas de problme */
}

/* -------- Calcule la contribution d'une couche au signal EXAFS --------\
 |   Renvoie 0 si tout s'est bien pass, != 0 sinon                      |
 |   AJOUTE la contribution  exafs, pour les X de exafs                 |
 | --> ne multiplie PAS par S0_2 (multiplication globale  la fin)       |
 | --> ne pondre PAS par k^n (car global  la fin - mais douteux :  voir |
 |                                                                       |
 | couche : la couche pour laquelle on veut la contribution              |
 | exafs  : le spectre dans lequel le signal EXAFS est enregistr        |
 \---------------------------------------------------------------------- */

static int exafs_1_couche(Couche *couche, Graphe *exafs)
{
  unsigned long i;
  register double *x, *y;
  double delta_e;

  if ( (couche->b_amp == NULL) || (couche->c_amp == NULL) || (couche->d_amp == NULL)
    || (couche->amplitude == NULL) )
    return -1; /* Pas d'informations pour les amplitudes */

  if ( (couche->b_pha == NULL) || (couche->c_pha == NULL) || (couche->d_pha == NULL)
    || (couche->phase == NULL) )
    return -2; /* Pas d'informations pour les phases */

  if ( deltae_general ) /* Les corrections d'nergie sont gnrales  chaque couche */
    delta_e = VAR_prend_valeur( &(Delta_E) );
  else
    delta_e = VAR_prend_valeur( &(couche->Delta_E) );

  for (i = 0, x = exafs->x, y = exafs->y ; i < exafs->nombre_point ; i++, x++, y++)
  {
    double amplitude, phase, debye_waller;
    double k_reel = *x, k_p_n; /* En fonction de Delta_E */
    double longueur;

    k_reel *= *x;
    k_reel += 2 * MASSE_ELECTRON * EV_V_JOULE / (H_BARRE * H_BARRE * M_V_ANGSTROM * M_V_ANGSTROM) * delta_e;
    k_reel = sqrt(k_reel);
    if (k_reel == 0)
    {
      continue;
    }

    longueur = VAR_prend_valeur( &(couche->R) );

    amplitude  = SPL_tt_calc(couche->amplitude, k_reel, couche->b_amp, couche->c_amp, couche->d_amp);
    amplitude *= VAR_prend_valeur( &(couche->N) );
    amplitude /= longueur * longueur;
    amplitude /= k_reel;

    phase  = SPL_tt_calc(couche->phase, k_reel, couche->b_pha, couche->c_pha, couche->d_pha);
    phase += 2 * k_reel * longueur;   

    /* Insrer ici le calcul du libre parcours moyen, si ncessaire */

    k_p_n = k_reel * k_reel; /* k */
    debye_waller  = VAR_prend_valeur( &(couche->Debye_Waller) ); /* Contient dj sigma */
    debye_waller *= -2.0 * k_p_n;

    k_p_n *= k_reel; /* k */
    phase -= 4 * VAR_prend_valeur( &(couche->Cumul_2) ) / 3 * k_p_n;

    k_p_n *= k_reel; /* k^4 */
    debye_waller += 2 * VAR_prend_valeur( &(couche->Cumul_1) ) / 3 * k_p_n;

    k_p_n *= k_reel; /* k^5 */
    phase += 4 * VAR_prend_valeur( &(couche->Cumul_4) ) / 15 * k_p_n;

    k_p_n *= k_reel; /* k^6 */
    debye_waller -= 4 * VAR_prend_valeur( &(couche->Cumul_3) ) / 45 * k_p_n;

    amplitude *= exp(debye_waller);

    *y += amplitude * sin ( phase );
  }

  return 0; /* Tout s'est bien pass */
}

/* -------- Calcul du signal EXAFS pour une couche en un point -----------\
 |  Renvoie la valeur du signal EXAFS au point considr                  |
 |                                                                        |
 | k      = le point pour lequel il faut calculer le signal EXAFS         |
 | couche = la couche  utiliser pour faire le calcul                     |
 \----------------------------------------------------------------------- */

double calc_exafs_1point(double k, Couche *couche)
{
  double amplitude, phase, debye_waller;
  double k_p_n; /* En fonction de Delta_E */
  double longueur, delta_e;

  if (deltae_general)
    delta_e = VAR_prend_valeur( &(Delta_E) );
  else
    delta_e = VAR_prend_valeur( &(couche->Delta_E) );

  if ( delta_e ) /* Correction d'nergie  faire */
  {
    k *= k;
    k += 2 * MASSE_ELECTRON * EV_V_JOULE / (H_BARRE * H_BARRE * M_V_ANGSTROM * M_V_ANGSTROM) * delta_e;
    k  = sqrt(k);
  }
  if (k == 0)
    return 0;

  longueur = VAR_prend_valeur( &(couche->R) );

  amplitude  = SPL_tt_calc(couche->amplitude, k, couche->b_amp, couche->c_amp, couche->d_amp);
  amplitude *= VAR_prend_valeur( &(couche->N) );
  amplitude /= longueur * longueur;

  if (ponder_ajuste == 0) /* Pas de pondration */
    amplitude /= k;
  else if (ponder_ajuste == 2) /* Pondration par k */
    amplitude *= k;       /* car se simplifie avec le /k */

  phase  = SPL_tt_calc(couche->phase, k, couche->b_pha, couche->c_pha, couche->d_pha);
  phase += 2 * k * longueur;   

  /* Insrer ici le calcul du libre parcours moyen, si ncessaire */

  k_p_n = k * k; /* k */
  debye_waller  = VAR_prend_valeur( &(couche->Debye_Waller) ); /* Contient dj sigma */
  debye_waller *= -2.0 * k_p_n;

  if (ponder_ajuste == 3) /* Pondration par k ; multiplier par k car se simplifie avec le 1/k */
    amplitude *= k_p_n;

  k_p_n *= k; /* k */
  phase -= 4 * VAR_prend_valeur( &(couche->Cumul_2) ) / 3 * k_p_n;

  if (ponder_ajuste == 4)
    amplitude *= k_p_n; /* Pondration par k^4 */

  k_p_n *= k; /* k^4 */
  debye_waller += 2 * VAR_prend_valeur( &(couche->Cumul_1) ) / 3 * k_p_n;

  if (ponder_ajuste == 5) /* Pondration par k^5 */
    amplitude *= k_p_n;

  k_p_n *= k; /* k^5 */
  phase += 4 * VAR_prend_valeur( &(couche->Cumul_4) ) / 15 * k_p_n;

  if (ponder_ajuste == 6) /* Pondration par k^6 */
    amplitude *= k_p_n;

  k_p_n *= k; /* k^6 */
  debye_waller -= 4 * VAR_prend_valeur( &(couche->Cumul_3) ) / 45 * k_p_n;

  if (ponder_ajuste == 7) /* Pondration par k^7 */
    amplitude *= k_p_n;

  amplitude *= exp(debye_waller);

  return VAR_prend_valeur( &(S0_2) ) * amplitude * sin ( phase );
}

/* -------- Calcul de la drive partielle du signal EXAFS ----------------\
 |     par rapport  une variable donne                                   |
 |     en un point donn                                                   |
 |     pour une couche donne                                              |
 |                                                                         |
 | k       = le point pour lequel on calcule                               |
 | couche  = la couche pour laquelle on fait le calcul                     |
 | type    = le type de variables (#define dans couches.h)                 |
 | val_khi = valeur du signal EXAFS en ce point                            |
 \------------------------------------------------------------------------ */

double calc_derivee(double k, Couche *couche, int type, double val_khi)
{
  double n, r, lpm;
  double amplitude, phase;
  double val_derivee;
  double delta_e;
  double k_p_n;

  switch (type)
  {
  case VAR_SO2: /* variable globale -> val_khi : signal EXAFS total
		   on ajuste en ralit s0 directement             */
    val_derivee = val_khi / VAR_prend_valeur( &(S0_2) );
    break;

  case VAR_DELTAE_GENE:
    break;

  case VAR_GAMMA_GENE:
    break;

  case VAR_N:
    val_derivee = val_khi / VAR_prend_valeur( &(couche->N) );
    break;

  case VAR_R:
    if (deltae_general)
      delta_e = VAR_prend_valeur( &(Delta_E) );
    else if (couche)
      delta_e = VAR_prend_valeur( &(couche->Delta_E) );

    if ( delta_e ) /* Correction d'nergie  faire */
    {
      k *= k;
      k += 2 * MASSE_ELECTRON * EV_V_JOULE / (H_BARRE * H_BARRE * M_V_ANGSTROM * M_V_ANGSTROM) * delta_e;
      k  = sqrt(k);
    }

    r = VAR_prend_valeur( &(couche->R) );

    val_derivee = -1 / r;
    if (type_gamma != GENE_LPM_INCLUS) /* Le lpm n'est pas inclus dans l'amplitude */
    {
      /* val_derivee -= 1 / lpm */
    }

    k_p_n = k * k * k; /* k */

    phase  = 2 * k * r;
    phase += SPL_tt_calc(couche->phase, k, couche->b_pha, couche->c_pha, couche->d_pha);
    phase -= 4/3  * k_p_n * VAR_prend_valeur( &(couche->Cumul_2 ) );
    k_p_n *= k * k;
    phase += 4/15 * k_p_n * VAR_prend_valeur( &(couche->Cumul_4 ) );

    val_derivee += k / tan(phase);
    val_derivee *= 2 * val_khi; /* dkhi/dR = 2 * khi * ( k/tg(2kR + phase) - 1/R - 1/lambda(k) )
				             suppose que l'amplitude est quasi-indpendante de R,
					     mme si le libre parcours moyen y est inclus !       */
    break;

  case VAR_DEBYE: /* La grandeur que l'on ajuste est en fait sigma */
    if (deltae_general)
      delta_e = VAR_prend_valeur( &(Delta_E) );
    else if (couche)
      delta_e = VAR_prend_valeur( &(couche->Delta_E) );

    if ( delta_e ) /* Correction d'nergie  faire */
    {
      k *= k;
      k += 2 * MASSE_ELECTRON * EV_V_JOULE / (H_BARRE * H_BARRE * M_V_ANGSTROM * M_V_ANGSTROM) * delta_e;
      k  = sqrt(k);
    }

    val_derivee = - val_khi * 2 * k * k;    
    break;

  case VAR_CUMUL_1: /* Premire correction pour Debye-Waller */
    if (deltae_general)
      delta_e = VAR_prend_valeur( &(Delta_E) );
    else if (couche)
      delta_e = VAR_prend_valeur( &(couche->Delta_E) );

    if ( delta_e ) /* Correction d'nergie  faire */
    {
      k *= k;
      k += 2 * MASSE_ELECTRON * EV_V_JOULE / (H_BARRE * H_BARRE * M_V_ANGSTROM * M_V_ANGSTROM) * delta_e;
      k  = sqrt(k);
    }

    val_derivee  = 2 * val_khi * k * k * k * k;
    val_derivee /= 3;
    break;

  case VAR_CUMUL_2: /* Premire correction pour le dphasage */
    if (deltae_general)
      delta_e = VAR_prend_valeur( &(Delta_E) );
    else if (couche)
      delta_e = VAR_prend_valeur( &(couche->Delta_E) );

    if ( delta_e ) /* Correction d'nergie  faire */
    {
      k *= k;
      k += 2 * MASSE_ELECTRON * EV_V_JOULE / (H_BARRE * H_BARRE * M_V_ANGSTROM * M_V_ANGSTROM) * delta_e;
      k  = sqrt(k);
    }

    k_p_n = k * k * k; /* k */

    phase  = 2 * k * r;
    phase += SPL_tt_calc(couche->phase, k, couche->b_pha, couche->c_pha, couche->d_pha);
    phase -= 4/3  * k_p_n * VAR_prend_valeur( &(couche->Cumul_2 ) );
    val_derivee = -4 * val_khi * k_p_n / 3;

    k_p_n *= k * k;
    phase += 4/15 * k_p_n * VAR_prend_valeur( &(couche->Cumul_4 ) );

    val_derivee /= tan(phase);
    break;

  case VAR_CUMUL_3: /* Deuxime correction pour Debye-Waller */
    if (deltae_general)
      delta_e = VAR_prend_valeur( &(Delta_E) );
    else if (couche)
      delta_e = VAR_prend_valeur( &(couche->Delta_E) );

    if ( delta_e ) /* Correction d'nergie  faire */
    {
      k *= k;
      k += 2 * MASSE_ELECTRON * EV_V_JOULE / (H_BARRE * H_BARRE * M_V_ANGSTROM * M_V_ANGSTROM) * delta_e;
      k  = sqrt(k);
    }

    k_p_n = k * k * k;
    k_p_n *= k_p_n; /* k_p_n = k^6 */
    val_derivee = - 4 * val_khi * k_p_n / 45;
    break;

  case VAR_CUMUL_4: /* Deuxime correction pour la phase */
    if (deltae_general)
      delta_e = VAR_prend_valeur( &(Delta_E) );
    else if (couche)
      delta_e = VAR_prend_valeur( &(couche->Delta_E) );

    if ( delta_e ) /* Correction d'nergie  faire */
    {
      k *= k;
      k += 2 * MASSE_ELECTRON * EV_V_JOULE / (H_BARRE * H_BARRE * M_V_ANGSTROM * M_V_ANGSTROM) * delta_e;
      k  = sqrt(k);
    }

    k_p_n = k * k * k; /* k */

    phase  = 2 * k * r;
    phase += SPL_tt_calc(couche->phase, k, couche->b_pha, couche->c_pha, couche->d_pha);
    phase -= 4/3  * k_p_n * VAR_prend_valeur( &(couche->Cumul_2 ) );
    k_p_n *= k * k; /* k_p_n = k^5 */
    phase += 4/15 * k_p_n * VAR_prend_valeur( &(couche->Cumul_4 ) );

    val_derivee  = 4/15 * val_khi * k_p_n;
    val_derivee /= tan(phase);
    break;

  case VAR_DELTAE: /* Un Delta E par couche si on est de ce type */
    break;

  case VAR_GAMMA:  /* Un gamma par couche si on est de ce type */
    break;
  }

  return val_derivee; /* Valeur de la drive en ce point */
}

/* -------- Calcul de l'erreur entre le thorique et l'ajust -------------\
 |  Renvoie l'erreur > 0, un code <0 si une erreur est survenue            |
 |  Suppose que le graphe de rfrence est tri suivant les X croissants   |
 |            et que les X convenables sont les X de l'ajust              |
 |                                                                         |
 | infos : les infos sur le modle utilis pour la minimisation            |
 \------------------------------------------------------------------------ */

double calc_erreur(Infos_ncouches *infos)
{
  unsigned long i = 0;
  double erreur = 0;
  register double *x_lu, *y_ref, *y_aj;

  if (infos == NULL)
    return -1;
  if (infos->reference == NULL)
    return -2;

  /* On passe les points de la rfrence qui ne servent pas  l'ajustement */

  x_lu  = infos->reference->x;
  y_ref = infos->reference->y;
  while ( *x_lu < kmin_ajuste )
  {
    x_lu++; y_ref++;
  }

  for (i = 0, y_aj = infos->ajuste->y;
       i < infos->ajuste->nombre_point ;
       i++, y_aj++, y_ref++)
  {
    double delta = *y_ref - *y_aj;

    delta *= delta;
/*    delta /= (*y_ref * *y_ref); */
    erreur += delta;
  }    

  return erreur;
}

/* -------- Calcul de la diffrence entre le thorique et l'ajust --------\
 |  Renvoie : l'erreur si tout s'est bien pass, un code <0 sinon          |
 |  Suppose que le graphe de rfrence est tri suivant les X croissants   |
 |            et que les X convenables sont les X de l'ajust              |
 |                                                                         |
 | infos : les infos sur le modle utilis pour la minimisation            |
 \------------------------------------------------------------------------ */

double calc_residu(Infos_ncouches *infos)
{
  unsigned long i = 0;
  double erreur = 0;
  register double *x_lu, *y_ref, *y_aj;

  if (infos == NULL)
    return -1;

  if ( (infos->reference == NULL) || (infos->ajuste == NULL) )
    return  -2; /* Pas de graphes pour travailler ! */

  x_lu  = infos->reference->x;
  y_ref = infos->reference->y;
  while ( *x_lu < kmin_ajuste )
  {
    x_lu++; y_ref++;
  }

  if ( infos->residu ) /* On calcule bien le spectre rsidu */
  {
    register double *y_residu;

    for (i = 0, y_aj = infos->ajuste->y, y_residu = infos->residu->y ;
	 i < infos->ajuste->nombre_point ;
	 i++, y_aj++, y_ref++, y_residu++)
    {
      double delta = *y_ref - *y_aj;
      *y_residu = delta;
      
      delta *= delta;
/*      delta /= (*y_ref * *y_ref); */
      erreur += delta;
    }

    GRF_bornes(infos->residu, ORDONNE, DESORDONNE); /* Nouvelles bornes */
  }
  else /* On calcule juste l'erreur */
  {
    for (i = 0, y_aj = infos->ajuste->y;
	 i < infos->ajuste->nombre_point ;
	 i++, y_aj++, y_ref++)
    {
      double delta = *y_ref - *y_aj;

      delta *= delta;
/*      delta /= (*y_ref * *y_ref); */
      erreur += delta;
    }    
  }

  return erreur;
}
